# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemSystemSetupWidget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(570, 883)
        Form.setMinimumSize(QtCore.QSize(0, 450))
        font = QtGui.QFont()
        font.setPointSize(10)
        Form.setFont(font)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 4, 0, 1, 3)
        self.comboBox_configuration = QtWidgets.QComboBox(Form)
        self.comboBox_configuration.setObjectName("comboBox_configuration")
        self.gridLayout.addWidget(self.comboBox_configuration, 0, 1, 1, 1)
        self.pushButton_connect_to_microscope = QtWidgets.QPushButton(Form)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.pushButton_connect_to_microscope.setFont(font)
        self.pushButton_connect_to_microscope.setObjectName("pushButton_connect_to_microscope")
        self.gridLayout.addWidget(self.pushButton_connect_to_microscope, 1, 0, 1, 3)
        self.toolButton_import_configuration = QtWidgets.QToolButton(Form)
        self.toolButton_import_configuration.setObjectName("toolButton_import_configuration")
        self.gridLayout.addWidget(self.toolButton_import_configuration, 0, 2, 1, 1)
        self.pushButton_apply_configuration = QtWidgets.QPushButton(Form)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.pushButton_apply_configuration.setFont(font)
        self.pushButton_apply_configuration.setObjectName("pushButton_apply_configuration")
        self.gridLayout.addWidget(self.pushButton_apply_configuration, 3, 0, 1, 3)
        self.label_configuration = QtWidgets.QLabel(Form)
        self.label_configuration.setObjectName("label_configuration")
        self.gridLayout.addWidget(self.label_configuration, 0, 0, 1, 1)
        self.label_connection_information = QtWidgets.QLabel(Form)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.label_connection_information.setFont(font)
        self.label_connection_information.setObjectName("label_connection_information")
        self.gridLayout.addWidget(self.label_connection_information, 5, 0, 1, 3)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.pushButton_connect_to_microscope.setText(_translate("Form", "Connect to Microscope"))
        self.toolButton_import_configuration.setText(_translate("Form", "..."))
        self.pushButton_apply_configuration.setText(_translate("Form", "Apply Microscope Configuration"))
        self.label_configuration.setText(_translate("Form", "Configuration"))
        self.label_connection_information.setText(_translate("Form", "No Connected"))



import matplotlib
import matplotlib.pyplot as plt

from fibsem import acquire, utils
from fibsem.structures import BeamType
import logging

matplotlib.use('TkAgg', force=True) # Activate 'agg' backend for off-screen plotting.


"""
This script will take an image with the electron beam, an image with the ion beam, and an image with both beams. 
The images are then displayed in a matplotlib figure.

The settings for images are stored in the settings.image struct, and can be modified before taking an image.

For more detail on the settings, see the documentation for the ImageSettings class.

"""

def main():
    
    # connect to the microscope
    microscope, settings = utils.setup_session(manufacturer="Demo", ip_address="localhost")

    # info about ImageSettings
    logging.info(f"\nAcquiring Images Example:")
    logging.info(f"The current image settings are: \n{settings.image}")

    # take an image with the electron beam
    settings.image.beam_type = BeamType.ELECTRON
    eb_image = acquire.new_image(microscope, settings.image)

    # take an image with the ion beam
    settings.image.beam_type = BeamType.ION
    ib_image = acquire.new_image(microscope, settings.image)

    # take an image with both beams with increased hfw
    settings.image.hfw = 400e-6       
    ref_eb_image, ref_ib_image = acquire.take_reference_images(microscope, settings.image)

    # show images

    fig, ax = plt.subplots(2, 2, figsize=(10, 7))
    ax[0][0].imshow(eb_image.data, cmap="gray")
    ax[0][0].set_title("Electron Image 01")
    ax[0][1].imshow(ib_image.data, cmap="gray")
    ax[0][1].set_title("Ion Image 01")
    ax[1][0].imshow(ref_eb_image.data, cmap="gray")
    ax[1][0].set_title("Electron Image 02 (Reference)")
    ax[1][1].imshow(ref_ib_image.data, cmap="gray")
    ax[1][1].set_title("Ion Image 02 (Reference)")
    plt.show()


if __name__ == "__main__":
    main()



import glob
import logging
import os
import re
import shutil
import json
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from PIL import Image
from copy import deepcopy

from fibsem import config as cfg
from fibsem import utils
from fibsem.detection.detection import DetectedFeatures
from fibsem.structures import FibsemImage, FibsemImageMetadata, Point

def decode_segmap(image, nc=3):

    """ Decode segmentation class mask into an RGB image mask"""

    # 0=background, 1=lamella, 2= needle
    label_colors = np.array([(0, 0, 0),
                                (255, 0, 0),
                                (0, 255, 0)])

    # pre-allocate r, g, b channels as zero
    r = np.zeros_like(image, dtype=np.uint8)
    g = np.zeros_like(image, dtype=np.uint8)
    b = np.zeros_like(image, dtype=np.uint8)

    # apply the class label colours to each pixel
    for l in range(0, nc):
        idx = image == l
        r[idx] = label_colors[l, 0]
        g[idx] = label_colors[l, 1]
        b[idx] = label_colors[l, 2]

    # stack rgb channels to form an image
    rgb_mask = np.stack([r, g, b], axis=2)
    return rgb_mask

def coordinate_distance(p1: Point, p2: Point):
    """Calculate the distance between two points in each coordinate"""

    return p2.x - p1.x, p2.y - p1.y

def scale_pixel_coordinates(px: Point, from_image: FibsemImage, to_image: FibsemImage) -> Point:
    """Scale the pixel coordinate from one image to another"""

    invariant_pt = get_scale_invariant_coordinates(px, from_image.data.shape)

    scaled_px = scale_coordinate_to_image(invariant_pt, to_image.data.shape)

    return scaled_px


def get_scale_invariant_coordinates(point: Point, shape: tuple) -> Point:
    """Convert the point coordinates from image coordinates to scale invariant coordinates"""
    scaled_pt = Point(x=point.x / shape[1], y=point.y / shape[0])

    return scaled_pt


def scale_coordinate_to_image(point: Point, shape: tuple) -> Point:
    """Scale invariant coordinates to image shape"""
    scaled_pt = Point(x=int(point.x * shape[1]), y=int(point.y * shape[0]))

    return scaled_pt

def parse_metadata(filename):

    # FIB meta data key is 34682, comes as a string
    img = Image.open(filename)
    img_metadata = img.tag[34682][0]

    # parse metadata
    parsed_metadata = img_metadata.split("\r\n")

    metadata_dict = {}
    for item in parsed_metadata:

        if item == "":
            # skip blank lines
            pass
        elif re.match(r"\[(.*?)\]", item):
            # find category, dont add to dict
            category = item
        else:
            # meta data point
            datum = item.split("=")

            # save to dictionary
            metadata_dict[category + "." + datum[0]] = datum[1]

    # add filename to metadata
    metadata_dict["filename"] = filename

    # convert to pandas df
    df = pd.DataFrame.from_dict(metadata_dict, orient="index").T

    return df

# TODO: add experiment, method
# TODO: migrate this to fibsem.db
# filename should match the same filename that is used for feature detection logging -> can be associated
def save_feature_data_to_csv(det: DetectedFeatures, features: list[dict], filename: str):
    """Save the feature data to a csv file. Includes saving the image and mask to disk. 
    All data is saved at the cfg.DATA_ML_PATH location. This can be configured in the config.py file."""

    # save image
    image = det.fibsem_image
    filename = os.path.join(cfg.DATA_ML_PATH, f"{filename}")
    image.save(filename) # type: ignore 
    logging.debug(f"Saved detection image to {filename}")

    # save mask to disk
    os.makedirs(os.path.join(cfg.DATA_ML_PATH, "mask"), exist_ok=True)
    mask_fname = os.path.join(cfg.DATA_ML_PATH, "mask", os.path.basename(filename))
    mask_fname = Path(mask_fname).with_suffix(".tif")
    im = Image.fromarray(det.mask) 
    im.save(mask_fname)
    logging.debug(f"Saved detection mask to {mask_fname}")

    # convert to csv format
    csv_data = []
    for fd in features:

        dat = deepcopy(fd)
        dat["px.x"] = fd["px"]["x"]
        dat["px.y"] = fd["px"]["y"]
        dat["dpx.x"] = fd["dpx"]["x"]
        dat["dpx.y"] = fd["dpx"]["y"]
        dat["dm.x"] = fd["dm"]["x"]
        dat["dm.y"] = fd["dm"]["y"]
        
        del dat["px"]
        del dat["dpx"]
        del dat["dm"]

        csv_data.append(dat)
    logging.debug(f"Converted {len(csv_data)} features to csv format")
    
    df = pd.DataFrame(csv_data)
    
    # save the dataframe to a csv file, append if the file already exists
    DATAFRAME_PATH = os.path.join(cfg.DATA_ML_PATH, "data.csv")
    if os.path.exists(DATAFRAME_PATH):
        df_tmp = pd.read_csv(DATAFRAME_PATH)
        df = pd.concat([df_tmp, df], axis=0, ignore_index=True)
    df.to_csv(DATAFRAME_PATH, index=False)
    logging.debug(f"Saved feature data to {DATAFRAME_PATH}")





import os
from tqdm import tqdm
import numpy as np
import matplotlib.pyplot as plt
import torch
import torch.nn.functional as F

from huggingface_hub import HfApi, hf_hub_download
from fibsem import config as cfg

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
from fibsem.segmentation.config import CLASS_COLORS, CLASS_COLORS_RGB, CLASS_LABELS


# helper functions
def decode_output(output):
    """decodes the output of segmentation model to RGB mask"""
    output = F.softmax(output, dim=1)
    mask = torch.argmax(output, dim=1).detach().cpu().numpy()
    return mask


def decode_segmap(image, nc=5):

    """
    Decode segmentation class mask into an RGB image mask
    ref: https://learnopencv.com/pytorch-for-beginners-semantic-segmentation-using-torchvision/
    """

    # 0=background, 1=lamella, 2= needle
    label_colors = np.zeros((8, 3))
    label_colors[1, :] = (255, 0, 0)
    label_colors[2, :] = (0, 255, 0)
    label_colors[3, :] = (0, 255, 255)
    label_colors[4, :] = (255, 255, 0)    
    label_colors[5, :] = (255, 0, 255)
    label_colors[6, :] = (0, 0, 255)
    label_colors[7, :] = (128, 0, 0) # TODO: convert this to use CLASS_COLORS

    # pre-allocate r, g, b channels as zero
    r = np.zeros_like(image, dtype=np.uint8)
    g = np.zeros_like(image, dtype=np.uint8)
    b = np.zeros_like(image, dtype=np.uint8)

    # TODO: make this more efficient
    # apply the class label colours to each pixel
    for class_idx in range(1, nc+1):
        idx = image == class_idx
        # class_idx = class_idx % len(label_colors)
        r[idx] = label_colors[class_idx, 0]
        g[idx] = label_colors[class_idx, 1]
        b[idx] = label_colors[class_idx, 2]

    # stack rgb channels to form an image
    rgb_mask = np.stack([r, g, b], axis=-1).squeeze()
    return rgb_mask
   
def decode_segmap_v2(image, colormap: list[tuple] = None) -> np.ndarray:
    """
    Decode segmentation class mask into an RGB image mask
    ref: https://learnopencv.com/pytorch-for-beginners-semantic-segmentation-using-torchvision/
    """

    if colormap is None:
        from fibsem.segmentation.config import CLASS_COLORS_RGB
        colormap = CLASS_COLORS_RGB

    # convert class masks to rgb values
    rgb_mask = np.zeros((image.shape[0], image.shape[1], 3), dtype=np.uint8)
    unique_labels = np.unique(image)

    for class_idx in unique_labels:
        rgb_mask[image == class_idx] = colormap[class_idx]

    return rgb_mask

def show_img_and_mask(imgs, gts, mask, title="Image, Ground Truth and Mask"):
    """Show a plot of the image, ground truth mask, and predicted mask"""
    n_imgs = len(imgs)
    imgs = imgs.cpu().detach()

    fig, ax = plt.subplots(n_imgs, 3, figsize=(8, 6))
    fig.suptitle(title)

    for i in range(len(imgs)):

        img = imgs[i].permute(1, 2, 0).squeeze()
        gt = decode_segmap(gts[i].permute(1, 2, 0).squeeze())  # convert to rgb mask

        ax[0].imshow(img, cmap="gray")
        ax[1].imshow(gt)
        ax[1].set_title("Ground Truth")
        ax[2].imshow(mask)
        ax[2].set_title("Predicted Mask")

    plt.show()


def show_memory_usage():
    """Show total, reserved and allocated gpu memory"""
    t = torch.cuda.get_device_properties(0).total_memory
    r = torch.cuda.memory_reserved(0)
    a = torch.cuda.memory_allocated(0)
    f = r - a  # free inside reserved
    if r == 0:
        r = 0.00001
    print("GPU memory", t, r, a, f, f"{f/r:.3}")


def show_values(ten):
    """Show tensor statistics for debugging"""
    unq = np.unique(ten.detach().cpu().numpy())
    print(ten.shape, ten.min(), ten.max(), ten.mean(), ten.std(), unq)


def validate_config(config:dict):
    if "data_paths" not in config:
        raise ValueError("data_path is missing. Should point to path containing labelled images.")
    else:
        for path in config["data_paths"]:
            if not os.path.exists(path):
                raise ValueError(f"{path} directory does not exist. (data_path)")
    if "label_paths" not in config:
        raise ValueError("label_path is missing. Should point to path containing labelled images.")
    else:
        for path in config["label_paths"]:
            if not os.path.exists(path):
                raise ValueError(f"{path} directory does not exist. (label_path)")
    if "save_path" not in config:
        raise ValueError("save_path is missing. Should point to save the model.")
    else:
        path = config["save_path"]
        os.makedirs(path, exist_ok=True)
    if "wandb" not in config:
        raise ValueError("wandb is missing. Used to enable/disable wandb logging in training loop. Should be a boolean value.")
    else:
        val = config["wandb"]
        if type(val) != bool:
            raise TypeError(f"{val} is not a boolean (True/False). (wandb)")
    if "checkpoint" not in config:
        raise ValueError("checkpoint is missing. Either a path leading to the desired saved model, or None value.")
    else:
        path = config["checkpoint"]
        if path is not None and not os.path.exists(path):
            raise ValueError(f"{path} directory does not exist. (checkpoint)")
    if "encoder" not in config:
        raise ValueError("encoder is missing. Used to specify which model architecture to use. Default is resnet18.")
    else:
        val = config["encoder"]
        if type(val) != str:
            raise TypeError(f"{val} must be a string. (encoder)")
        elif val not in unet_encoders:
            raise ValueError(f"{val} not a valid encoder. Check readme for full list. (encoder)")
    if "epochs" not in config:
        raise ValueError("epochs is missing. Integer value used to determine number of epochs model trains for.")
    else:
        val = config["epochs"]
        if type(val) != int or val <= 0:
            raise TypeError(f"{val} is not a positive integer. (epochs)")  
    if "batch_size" not in config:
        raise ValueError("batch_size is missing. Integer value used to determine batch size of dataset.")
    else:
        val = config["batch_size"]
        if type(val) != int or val <= 0:
            raise TypeError(f"{val} is not a positive integer. (batch_size)")    
    if "num_classes" not in config:
        raise ValueError("num_classes is missing. Integer value used to determine number of classes model classifies.")
    else:
        val = config["num_classes"]
        if type(val) != int or val <= 0:
            raise TypeError(f"{val} is not a positive integer. (num_classes)")  
    if "lr" not in config:
        raise ValueError("lr is missing. Float value indicating the learning rate of the model.")
    else:
        val = config["lr"]
        if type(val) == float:
            if val <= 0:
                raise ValueError(f"{val} must be a positive float value (lr).")
        else:
            raise TypeError(f"{val} is not a float. (lr)")  
    if "wandb_project" not in config:
        raise ValueError("wandb_project is missing. String indicating the wandb project title for login.")
    else:
        val = config["wandb_project"]
        if type(val) != str:
            raise TypeError(f"{val} is not a string. (wandb_project)")
    if "wandb_entity" not in config:
        raise ValueError("wandb_entity is missing. String indicating the wandb login credentials.")
    else:
        val = config["wandb_entity"]
        if type(val) != str:
            raise TypeError(f"{val} is not a string. (wandb_project)")
    return
    
        
# All UNet encoders that work with Imagenet weights
unet_encoders = [
    "resnet18",
    "resnet34",
    "resnet50",
    "resnet101",
    "resnet152",
    "resnext50_32x4d",
    "resnext101_32x16d",
    "resnext101_32x32d",
    "resnext101_32x48d",
    "dpn68",
    "dpn98",
    "dpn131",
    "vgg11",
    "vgg11_bn",
    "vgg13",
    "vgg13_bn",
    "vgg16",
    "vgg16_bn",
    "vgg19",
    "vgg19_bn",
    "senet154",
    "se_resnet50",
    "se_resnet101",
    "se_resnet152",
    "se_resnext50_32x4d",
    "se_resnext101_32x4d",
    "densenet121",
    "densenet169",
    "densenet201",
    "densenet161",
    "efficientnet-b0",
    "efficientnet-b1",
    "efficientnet-b2",
    "efficientnet-b3",
    "efficientnet-b4",
    "efficientnet-b5",
    "efficientnet-b6",
    "efficientnet-b7",
    "mobilenet_v2",
    "efficientnet-b0",
    "xception"
]
        

def plot_segmentations(images: list[np.ndarray], masks: list[np.ndarray], 
    alpha=0.5, legend: bool = True, show: bool = True) -> plt.Figure:
    """Plot the image and mask overlaid with a class legend."""
    
    if not isinstance(images, list):
        images = [images]
    if not isinstance(masks, list):
        masks = [masks]

    if len(images) != len(masks):
        raise ValueError("images and masks must be the same length")
    
    n_cols = len(images)
    fig, ax = plt.subplots(1, len(images), figsize=(10*n_cols/2, 10*n_cols/2))
    for i, (image, mask) in enumerate(zip(images, masks)):

        # convert to rgb mask        
        rgb = decode_segmap_v2(mask)

        if len(images) == 1:
            axes = ax
        else:
            axes = ax[i]
        # plot
        axes.imshow(image.data, cmap='gray')
        axes.imshow(rgb, alpha=0.4)

        # filter legend to only include classes present in mask
        class_ids = np.unique(mask)
        
        colors, labels = [], []
        for idx in class_ids:
            colors.append(CLASS_COLORS[idx])
            labels.append(CLASS_LABELS[idx])

        # Create a patch for each class color
        patches = [mpatches.Patch(color=color, label=label) 
                for color, label in zip(colors, labels)]

        # Add the patches to the legend
        if legend:
            axes.legend(handles=patches, loc="best", prop={'size': 6})
    
    if show:
        plt.show()

    return fig

## Huggingface Utils

def list_available_checkpoints():
    api = HfApi()
    files = api.list_repo_files(cfg.HUGGINFACE_REPO)
    checkpoints = []
    for file in files:
        if file.endswith(".pt") and "archive" not in file:
            checkpoints.append(file)

    return checkpoints

def download_checkpoint(checkpoint: str):
    if os.path.exists(checkpoint):
        checkpoint = checkpoint
    else:
        checkpoint = hf_hub_download(repo_id=cfg.HUGGINFACE_REPO, filename=checkpoint)
    return checkpoint


from fibsem import utils
from fibsem.structures import  FibsemStagePosition
import numpy as np
import logging


"""
This script demonstrates how to get the current stage position, and how to move the stage to a new position.

The basic movement methods are absolute_move and relative_move. 
- Relative move moves the stage by a certain amount in the current coordinate system.
- Absolute move moves the stage to a new position in the absolute coordinate system. 

This script will move the stage by 20um in the x direction (relative move), and then move back to the original position (absolute move).

Additional movement methods are available in the core api:
- Stable Move: the stage moves along the sample plane, accounting for stage tilt, and shuttle pre-tilt
- Vertical Move: the stage moves vertically in the chamber, regardless of tilt orientation

"""

def main():

    # connect to microscope
    microscope, settings = utils.setup_session(manufacturer="Demo", ip_address="localhost")
    
        # info about ImageSettings
    logging.info("---------------------------------- Current Position ----------------------------------\n")

    # get current position
    intial_position = microscope.get_stage_position()
    logging.info(f"\nStage Movement Example:")
    logging.info(f"Current stage position: {intial_position}")
    

    logging.info("\n---------------------------------- Relative Movement ----------------------------------\n")

    #### Moving to a relative position ####
    relative_move = FibsemStagePosition(x=20e-6,            # metres
                                        y=0,                # metres
                                        z=0.0,              # metres
                                        r=np.deg2rad(0),    # radians
                                        t=np.deg2rad(0))    # radians
    
    input(f"Press Enter to move by: {relative_move} (Relative)")
    
    # move by relative position    
    microscope.move_stage_relative(relative_move)
    current_position = microscope.get_stage_position()
    logging.info(f"After move stage position: {current_position}")


    logging.info("\n---------------------------------- Absolute Movement ----------------------------------\n")

    #### Moving to an absolute position ####
    stage_position = intial_position # move back to initial position

    # uncomment this if you want to move to a different position 
    # be careful to define a safe position to move too
    # relative_move = FibsemStagePosition(x=0,                # metres
    #                                     y=0,                # metres
    #                                 z=0.0,                  # metres
    #                                     r=np.deg2rad(0),    # radians
    #                                     t=np.deg2rad(0))    # radians

    input(f"Press Enter to move to: {stage_position} (Absolute)")

    # move to absolute position
    microscope.move_stage_absolute(stage_position) 
    current_position = microscope.get_stage_position()
    logging.info(f"After move stage position: {current_position}")


    logging.info("---------------------------------- End Example ----------------------------------")
   

if __name__ == "__main__":
    main()


import argparse
import os
import random
from datetime import datetime

import numpy as np
import segmentation_models_pytorch as smp
import torch

try:
    import wandb
except:
    pass

import yaml
from tqdm import tqdm

from fibsem.segmentation import dataset, utils

from skimage.color import gray2rgb
from skimage.util import img_as_ubyte


def _convert_checkpoint_format(checkpoint: str, encoder:str, nc: int, output_filename: str):
    """Converts a checkpoint from the old format to the new format"""
    import torch
    from huggingface_hub import hf_hub_download

    REPO_ID = "patrickcleeve/autolamella"
    checkpoint = hf_hub_download(repo_id=REPO_ID, filename=checkpoint)

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    checkpoint_state = torch.load(checkpoint, map_location=device)
    state_dict = {"checkpoint": checkpoint_state, "encoder": encoder, "nc": nc}

    # save
    torch.save(state_dict, output_filename)
    print(f"Saved as {output_filename}")

def _create_wandb_image(img, gt, pred, caption):
    
    img = img_as_ubyte(gray2rgb(img)) 
    gt = utils.decode_segmap_v2(gt[0]) 
    pred = utils.decode_segmap_v2(pred[0])

    return wandb.Image(
        np.hstack([img, gt, pred]), caption=caption)


# TODO: update save model to new checkpoint format
def save_model(save_dir, model, epoch):
    """Helper function for saving the model based on current time and epoch"""

    # dt_string = datetime.now().strftime("%d_%m_%Y_%H_%M_%S") + f"_n{epoch+1:02d}"
    model_save_file = os.path.join(save_dir, f"model{epoch:02d}.pt")
    torch.save(model.state_dict(), model_save_file)

    print(f"Model saved to {model_save_file}")

def save_model_v2(save_dir, model, epoch, encoder, nc):
    # save model with all necessary information

    # datetime as YYYYMMDD
    dt_string = datetime.now().strftime("%Y%m%d")
    checkpoint_name = os.path.join(save_dir, f"model-{dt_string}-{epoch:02d}.pt")
    checkpoint_state = model.state_dict()

    state_dict = {"checkpoint": checkpoint_state, "encoder": encoder, "nc": nc}

    # save
    torch.save(state_dict, checkpoint_name)
    print(f"Checkpoint saved as {checkpoint_name}")

def train(model, device, data_loader, criterion, optimizer, WANDB, log_freq: int = 32):
    data_loader = tqdm(data_loader)
    train_loss = 0

    for i, (images, masks) in enumerate(data_loader):
        # set model to training mode
        model.train()

        # move img and mask to device, reshape mask
        images = images.to(device)
        masks = masks.type(torch.LongTensor)
        masks = masks.reshape(
            masks.shape[0], masks.shape[2], masks.shape[3]
        )  # remove channel dim
        masks = masks.to(device)

        # forward pass
        outputs = model(images).type(torch.FloatTensor).to(device)
        loss = criterion(outputs, masks)

        # backwards pass
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        # evaluation
        train_loss += loss.item()

        data_loader.set_description(f"Train Loss: {loss.item():.04f}")


        if WANDB and i % log_freq == 0: 
            
            # TODO: this is really inefficient, re-running the model on the same image, just use existing outputs
            idx = random.choice(np.arange(0, images.shape[0]))
            output = model(images[idx][None, :, :, :])
            output_mask = utils.decode_output(output)

            img_base = images[idx].detach().cpu().squeeze().numpy()
            gt_base = masks[idx].detach().cpu()[:, :, None].permute(2, 0, 1).numpy()
            stack = _create_wandb_image(img_base, gt_base, output_mask, "Train Image (Raw, GT, Pred)")
            wandb.log({"train_loss": loss.item(), "train_image": stack})

    return train_loss


def validate(model, device, data_loader, criterion, WANDB, log_freq: int = 16):
    val_loader = tqdm(data_loader)
    val_loss = 0

    for i, (images, masks) in enumerate(val_loader):

        model.eval()

        # move img and mask to device, reshape mask
        images = images.to(device)
        masks = masks.type(torch.LongTensor)
        masks = masks.reshape(
            masks.shape[0], masks.shape[2], masks.shape[3]
        )  # remove channel dim
        masks = masks.to(device)

        # forward pass
        outputs = model(images).type(torch.FloatTensor).to(device)
        loss = criterion(outputs, masks)

        val_loss += loss.item()

        val_loader.set_description(f"Val Loss: {loss.item():.04f}")


        if WANDB and i % log_freq == 0:
            
            # select random image from batch
            # output = model(images[0][None, :, :, :])
            output_mask = utils.decode_output(outputs)

            img_base = images[0].detach().cpu().squeeze().numpy()
            gt_base = masks[0].detach().cpu()[:, :, None].permute(2, 0, 1).numpy()

            stack = _create_wandb_image(img_base, gt_base, output_mask, "Val Image (Raw, GT, Pred)")
            wandb.log({"val_loss": loss.item(), "val_image": stack})

    return val_loss

# initialise loss function and optimizer
def multi_loss(pred, target) -> float:
    c_loss = torch.nn.CrossEntropyLoss()
    d_loss = smp.losses.DiceLoss(mode="multiclass")
    f_loss = smp.losses.FocalLoss(mode="multiclass")
    return c_loss(pred, target) + d_loss(pred, target) + f_loss(pred, target)


def train_model(
    model,
    device,
    optimizer,
    train_data_loader,
    val_data_loader,
    config: dict, 
):
    """Helper function for training the model"""

    epochs=config["epochs"]
    WANDB=config["wandb"]
    TRAIN_LOG_FREQ=config.get("train_log_freq", 32)
    VAL_LOG_FREQ=config.get("val_log_freq", 16)

    criterion = multi_loss

    total_steps = len(train_data_loader)
    print(f"{epochs} epochs, {total_steps} total_steps per epoch")

    train_losses = []
    val_losses = []

    # training loop
    for epoch in range(epochs):
        print(f"------- Epoch {epoch+1} of {epochs}  --------")

        train_loss = train(model, device, train_data_loader, criterion, optimizer, WANDB, TRAIN_LOG_FREQ)
        val_loss = validate(model, device, val_data_loader, criterion, WANDB, VAL_LOG_FREQ)

        train_losses.append(train_loss / len(train_data_loader))
        val_losses.append(val_loss / len(val_data_loader))
        
        # get index of lowest loss
        print(f"MIN TRAIN LOSS at {train_losses.index(min(train_losses))}")
        print(f"MIN VAL LOSS at {val_losses.index(min(val_losses))}")

        # save_model(save_dir, model, epoch)
        save_model_v2(config["save_path"], model, epoch, config["encoder"], config["num_classes"])

    return model


def _setup_model(config: dict) -> tuple:
    # from smp
    model = smp.Unet(
        encoder_name=config["encoder"],
        encoder_weights="imagenet",
        in_channels=1,  # grayscale images
        classes=int(config["num_classes"]),  # background, needle, lamella
    )

    # Use gpu for training if available else use cpu
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    model.to(device)

    # load model checkpoint
    if config["checkpoint"]:
        model.load_state_dict(torch.load(config["checkpoint"], map_location=device))
        print(f"Checkpoint file {config['checkpoint']} loaded.")

    optimizer = torch.optim.Adam(model.parameters(), lr=config["lr"])

    return model, optimizer, device

def _setup_dataset(config:dict):

    train_data_loader, val_data_loader = dataset.preprocess_data(
        data_paths = config["data_paths"], 
        label_paths= config["label_paths"], 
        num_classes=config["num_classes"], 
        batch_size=config["batch_size"],
        val_split=config.get("split", 0.15),
        _validate_dataset=config.get("validate_dataset", True),
        apply_transforms=config.get("apply_transforms", False),
    )

    return train_data_loader, val_data_loader


def _setup_wandb(config:dict):
    # other parameters
    WANDB = config["wandb"]
    if WANDB:
        # weights and biases setup
        wandb.init(
            project=config["wandb_project"],
            entity=config["wandb_entity"],
            config=config,
        )

def main(config: dict):

    utils.validate_config(config)

    _setup_wandb(config)

    ################################## LOAD DATASET ##################################
    train_data_loader, val_data_loader = _setup_dataset(config)

    ################################## LOAD MODEL ##################################

    model, optimizer, device = _setup_model(config)
        
    ################################## TRAINING ##################################

    # train model
    model = train_model(
        model,
        device,
        optimizer,
        train_data_loader,
        val_data_loader,
        config=config,
    )

if __name__ == "__main__":
    # command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        help="the directory containing the config file to use",
        dest="config",
        action="store",
        default=os.path.join(os.path.join(os.path.dirname(__file__), "config.yml")),
    )
    args = parser.parse_args()
    config_dir = args.config

    # NOTE: Setup your config.yml file
    with open(config_dir, "r") as f:
        config = yaml.safe_load(f)

    main(config)    


import pytest

from fibsem.structures import MicroscopeState, BeamType, ImageSettings, FibsemImage, FibsemRectangle, BeamSettings, FibsemDetectorSettings, FibsemStagePosition

import datetime
# microscope state


# electron_beam, electron_detector, ion_beam, ion_detector are now optional


def test_microscope_state():

    state = MicroscopeState()

    state.to_dict()

    state.electron_beam = None
    state.electron_detector = None

    state.to_dict()

    state.ion_beam = None
    state.ion_detector = None

    state.to_dict()
    



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemMillingWidget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(608, 850)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.pushButton_run_milling = QtWidgets.QPushButton(Form)
        self.pushButton_run_milling.setObjectName("pushButton_run_milling")
        self.gridLayout.addWidget(self.pushButton_run_milling, 26, 0, 1, 2)
        self.frame = QtWidgets.QFrame(Form)
        self.frame.setMinimumSize(QtCore.QSize(0, 0))
        self.frame.setObjectName("frame")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.frame)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.comboBox_milling_stage = QtWidgets.QComboBox(self.frame)
        self.comboBox_milling_stage.setObjectName("comboBox_milling_stage")
        self.gridLayout_4.addWidget(self.comboBox_milling_stage, 2, 1, 1, 1)
        self.doubleSpinBox_centre_y = QtWidgets.QDoubleSpinBox(self.frame)
        self.doubleSpinBox_centre_y.setMinimum(-1e+16)
        self.doubleSpinBox_centre_y.setMaximum(1e+23)
        self.doubleSpinBox_centre_y.setSingleStep(0.1)
        self.doubleSpinBox_centre_y.setObjectName("doubleSpinBox_centre_y")
        self.gridLayout_4.addWidget(self.doubleSpinBox_centre_y, 22, 1, 1, 1)
        self.pushButton_remove_milling_stage = QtWidgets.QPushButton(self.frame)
        self.pushButton_remove_milling_stage.setObjectName("pushButton_remove_milling_stage")
        self.gridLayout_4.addWidget(self.pushButton_remove_milling_stage, 3, 1, 1, 1)
        self.label_centre_x = QtWidgets.QLabel(self.frame)
        self.label_centre_x.setObjectName("label_centre_x")
        self.gridLayout_4.addWidget(self.label_centre_x, 21, 0, 1, 1)
        self.label_hfw = QtWidgets.QLabel(self.frame)
        self.label_hfw.setObjectName("label_hfw")
        self.gridLayout_4.addWidget(self.label_hfw, 8, 0, 1, 1)
        self.checkBox_relative_move = QtWidgets.QCheckBox(self.frame)
        self.checkBox_relative_move.setChecked(True)
        self.checkBox_relative_move.setObjectName("checkBox_relative_move")
        self.gridLayout_4.addWidget(self.checkBox_relative_move, 27, 1, 1, 1)
        self.doubleSpinBox_dwell_time = QtWidgets.QDoubleSpinBox(self.frame)
        self.doubleSpinBox_dwell_time.setDecimals(4)
        self.doubleSpinBox_dwell_time.setMinimum(0.0)
        self.doubleSpinBox_dwell_time.setMaximum(4000000.0)
        self.doubleSpinBox_dwell_time.setSingleStep(0.01)
        self.doubleSpinBox_dwell_time.setProperty("value", 0.0)
        self.doubleSpinBox_dwell_time.setObjectName("doubleSpinBox_dwell_time")
        self.gridLayout_4.addWidget(self.doubleSpinBox_dwell_time, 11, 1, 1, 1)
        self.checkBox_live_update = QtWidgets.QCheckBox(self.frame)
        self.checkBox_live_update.setObjectName("checkBox_live_update")
        self.gridLayout_4.addWidget(self.checkBox_live_update, 27, 0, 1, 1)
        self.doubleSpinBox_spot_size = QtWidgets.QDoubleSpinBox(self.frame)
        self.doubleSpinBox_spot_size.setDecimals(4)
        self.doubleSpinBox_spot_size.setMinimum(0.0)
        self.doubleSpinBox_spot_size.setMaximum(100000.0)
        self.doubleSpinBox_spot_size.setSingleStep(0.01)
        self.doubleSpinBox_spot_size.setProperty("value", 0.0)
        self.doubleSpinBox_spot_size.setObjectName("doubleSpinBox_spot_size")
        self.gridLayout_4.addWidget(self.doubleSpinBox_spot_size, 12, 1, 1, 1)
        self.doubleSpinBox_rate = QtWidgets.QDoubleSpinBox(self.frame)
        self.doubleSpinBox_rate.setDecimals(4)
        self.doubleSpinBox_rate.setMinimum(0.0)
        self.doubleSpinBox_rate.setMaximum(100000.0)
        self.doubleSpinBox_rate.setSingleStep(0.01)
        self.doubleSpinBox_rate.setProperty("value", 0.0)
        self.doubleSpinBox_rate.setObjectName("doubleSpinBox_rate")
        self.gridLayout_4.addWidget(self.doubleSpinBox_rate, 10, 1, 1, 1)
        self.comboBox_patterns = QtWidgets.QComboBox(self.frame)
        self.comboBox_patterns.setObjectName("comboBox_patterns")
        self.gridLayout_4.addWidget(self.comboBox_patterns, 19, 1, 1, 1)
        self.label_centre_y = QtWidgets.QLabel(self.frame)
        self.label_centre_y.setObjectName("label_centre_y")
        self.gridLayout_4.addWidget(self.label_centre_y, 22, 0, 1, 1)
        self.label_voltage = QtWidgets.QLabel(self.frame)
        self.label_voltage.setObjectName("label_voltage")
        self.gridLayout_4.addWidget(self.label_voltage, 6, 0, 1, 1)
        self.doubleSpinBox_spacing = QtWidgets.QDoubleSpinBox(self.frame)
        self.doubleSpinBox_spacing.setProperty("value", 1.0)
        self.doubleSpinBox_spacing.setObjectName("doubleSpinBox_spacing")
        self.gridLayout_4.addWidget(self.doubleSpinBox_spacing, 13, 1, 1, 1)
        self.gridLayout_patterns = QtWidgets.QGridLayout()
        self.gridLayout_patterns.setObjectName("gridLayout_patterns")
        self.gridLayout_4.addLayout(self.gridLayout_patterns, 24, 0, 2, 2)
        self.line_3 = QtWidgets.QFrame(self.frame)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.gridLayout_4.addWidget(self.line_3, 4, 0, 1, 2)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_4.addItem(spacerItem, 26, 0, 1, 2)
        self.label_pattern_set = QtWidgets.QLabel(self.frame)
        self.label_pattern_set.setObjectName("label_pattern_set")
        self.gridLayout_4.addWidget(self.label_pattern_set, 19, 0, 1, 1)
        self.comboBox_application_file = QtWidgets.QComboBox(self.frame)
        self.comboBox_application_file.setObjectName("comboBox_application_file")
        self.gridLayout_4.addWidget(self.comboBox_application_file, 9, 1, 1, 1)
        self.label_milling_header = QtWidgets.QLabel(self.frame)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.label_milling_header.setFont(font)
        self.label_milling_header.setObjectName("label_milling_header")
        self.gridLayout_4.addWidget(self.label_milling_header, 5, 0, 1, 2)
        self.label_milling_instructions = QtWidgets.QLabel(self.frame)
        self.label_milling_instructions.setObjectName("label_milling_instructions")
        self.gridLayout_4.addWidget(self.label_milling_instructions, 29, 0, 1, 2)
        self.doubleSpinBox_centre_x = QtWidgets.QDoubleSpinBox(self.frame)
        self.doubleSpinBox_centre_x.setMinimum(-1e+28)
        self.doubleSpinBox_centre_x.setMaximum(1e+18)
        self.doubleSpinBox_centre_x.setSingleStep(0.1)
        self.doubleSpinBox_centre_x.setObjectName("doubleSpinBox_centre_x")
        self.gridLayout_4.addWidget(self.doubleSpinBox_centre_x, 21, 1, 1, 1)
        self.label_spot_size = QtWidgets.QLabel(self.frame)
        self.label_spot_size.setObjectName("label_spot_size")
        self.gridLayout_4.addWidget(self.label_spot_size, 12, 0, 1, 1)
        self.spinBox_voltage = QtWidgets.QSpinBox(self.frame)
        self.spinBox_voltage.setMaximum(1000000)
        self.spinBox_voltage.setProperty("value", 30000)
        self.spinBox_voltage.setObjectName("spinBox_voltage")
        self.gridLayout_4.addWidget(self.spinBox_voltage, 6, 1, 1, 1)
        self.label_preset = QtWidgets.QLabel(self.frame)
        self.label_preset.setObjectName("label_preset")
        self.gridLayout_4.addWidget(self.label_preset, 14, 0, 1, 1)
        self.label_application_file = QtWidgets.QLabel(self.frame)
        self.label_application_file.setObjectName("label_application_file")
        self.gridLayout_4.addWidget(self.label_application_file, 9, 0, 1, 1)
        self.label_dwell_time = QtWidgets.QLabel(self.frame)
        self.label_dwell_time.setObjectName("label_dwell_time")
        self.gridLayout_4.addWidget(self.label_dwell_time, 11, 0, 1, 1)
        self.label_milling_stage = QtWidgets.QLabel(self.frame)
        self.label_milling_stage.setObjectName("label_milling_stage")
        self.gridLayout_4.addWidget(self.label_milling_stage, 2, 0, 1, 1)
        self.doubleSpinBox_hfw = QtWidgets.QDoubleSpinBox(self.frame)
        self.doubleSpinBox_hfw.setEnabled(True)
        self.doubleSpinBox_hfw.setReadOnly(True)
        self.doubleSpinBox_hfw.setMinimum(10.0)
        self.doubleSpinBox_hfw.setMaximum(1000000000.0)
        self.doubleSpinBox_hfw.setProperty("value", 150.0)
        self.doubleSpinBox_hfw.setObjectName("doubleSpinBox_hfw")
        self.gridLayout_4.addWidget(self.doubleSpinBox_hfw, 8, 1, 1, 1)
        self.line_2 = QtWidgets.QFrame(self.frame)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.gridLayout_4.addWidget(self.line_2, 16, 0, 1, 2)
        self.doubleSpinBox_milling_current = QtWidgets.QDoubleSpinBox(self.frame)
        self.doubleSpinBox_milling_current.setObjectName("doubleSpinBox_milling_current")
        self.gridLayout_4.addWidget(self.doubleSpinBox_milling_current, 7, 1, 1, 1)
        self.comboBox_preset = QtWidgets.QComboBox(self.frame)
        self.comboBox_preset.setObjectName("comboBox_preset")
        self.gridLayout_4.addWidget(self.comboBox_preset, 14, 1, 1, 1)
        self.label_rate = QtWidgets.QLabel(self.frame)
        self.label_rate.setObjectName("label_rate")
        self.gridLayout_4.addWidget(self.label_rate, 10, 0, 1, 1)
        self.pushButton_update_pattern = QtWidgets.QPushButton(self.frame)
        self.pushButton_update_pattern.setObjectName("pushButton_update_pattern")
        self.gridLayout_4.addWidget(self.pushButton_update_pattern, 30, 0, 1, 2)
        self.label_milling_current = QtWidgets.QLabel(self.frame)
        self.label_milling_current.setObjectName("label_milling_current")
        self.gridLayout_4.addWidget(self.label_milling_current, 7, 0, 1, 1)
        self.label_patterns_header = QtWidgets.QLabel(self.frame)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.label_patterns_header.setFont(font)
        self.label_patterns_header.setObjectName("label_patterns_header")
        self.gridLayout_4.addWidget(self.label_patterns_header, 17, 0, 1, 2)
        self.label_spacing = QtWidgets.QLabel(self.frame)
        self.label_spacing.setObjectName("label_spacing")
        self.gridLayout_4.addWidget(self.label_spacing, 13, 0, 1, 1)
        self.pushButton_add_milling_stage = QtWidgets.QPushButton(self.frame)
        self.pushButton_add_milling_stage.setObjectName("pushButton_add_milling_stage")
        self.gridLayout_4.addWidget(self.pushButton_add_milling_stage, 3, 0, 1, 1)
        self.checkBox_show_milling_crosshair = QtWidgets.QCheckBox(self.frame)
        self.checkBox_show_milling_crosshair.setChecked(True)
        self.checkBox_show_milling_crosshair.setObjectName("checkBox_show_milling_crosshair")
        self.gridLayout_4.addWidget(self.checkBox_show_milling_crosshair, 28, 1, 1, 1)
        self.gridLayout.addWidget(self.frame, 16, 0, 1, 2)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem1, 29, 0, 1, 2)
        self.label_info = QtWidgets.QLabel(Form)
        self.label_info.setText("")
        self.label_info.setWordWrap(True)
        self.label_info.setObjectName("label_info")
        self.gridLayout.addWidget(self.label_info, 28, 0, 1, 2)
        self.line = QtWidgets.QFrame(Form)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.gridLayout.addWidget(self.line, 18, 0, 1, 2)
        self.progressBar_milling = QtWidgets.QProgressBar(Form)
        self.progressBar_milling.setProperty("value", 24)
        self.progressBar_milling.setObjectName("progressBar_milling")
        self.gridLayout.addWidget(self.progressBar_milling, 27, 0, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)
        Form.setTabOrder(self.frame, self.comboBox_milling_stage)
        Form.setTabOrder(self.comboBox_milling_stage, self.pushButton_add_milling_stage)
        Form.setTabOrder(self.pushButton_add_milling_stage, self.pushButton_remove_milling_stage)
        Form.setTabOrder(self.pushButton_remove_milling_stage, self.checkBox_live_update)
        Form.setTabOrder(self.checkBox_live_update, self.pushButton_update_pattern)
        Form.setTabOrder(self.pushButton_update_pattern, self.pushButton_run_milling)
        Form.setTabOrder(self.pushButton_run_milling, self.doubleSpinBox_hfw)
        Form.setTabOrder(self.doubleSpinBox_hfw, self.comboBox_application_file)
        Form.setTabOrder(self.comboBox_application_file, self.doubleSpinBox_rate)
        Form.setTabOrder(self.doubleSpinBox_rate, self.doubleSpinBox_dwell_time)
        Form.setTabOrder(self.doubleSpinBox_dwell_time, self.doubleSpinBox_spot_size)
        Form.setTabOrder(self.doubleSpinBox_spot_size, self.doubleSpinBox_spacing)
        Form.setTabOrder(self.doubleSpinBox_spacing, self.comboBox_preset)
        Form.setTabOrder(self.comboBox_preset, self.comboBox_patterns)
        Form.setTabOrder(self.comboBox_patterns, self.doubleSpinBox_centre_x)
        Form.setTabOrder(self.doubleSpinBox_centre_x, self.doubleSpinBox_centre_y)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.pushButton_run_milling.setText(_translate("Form", "Run Milling"))
        self.pushButton_remove_milling_stage.setText(_translate("Form", "Remove"))
        self.label_centre_x.setText(_translate("Form", "Centre X (um)"))
        self.label_hfw.setText(_translate("Form", "Horizontal Field Width (um)"))
        self.checkBox_relative_move.setText(_translate("Form", "Keep Relative Orientation"))
        self.checkBox_live_update.setText(_translate("Form", "Live Update"))
        self.label_centre_y.setText(_translate("Form", "Centre Y (um)"))
        self.label_voltage.setText(_translate("Form", "Voltage (V)"))
        self.label_pattern_set.setText(_translate("Form", "Pattern"))
        self.label_milling_header.setText(_translate("Form", "Milling"))
        self.label_milling_instructions.setText(_translate("Form", "Controls:"))
        self.label_spot_size.setText(_translate("Form", "Spot Size (um)"))
        self.label_preset.setText(_translate("Form", "Preset"))
        self.label_application_file.setText(_translate("Form", "Application File"))
        self.label_dwell_time.setText(_translate("Form", "Dwell Time (us)"))
        self.label_milling_stage.setText(_translate("Form", "Milling Stage"))
        self.label_rate.setText(_translate("Form", "Rate (mm3/A/s)"))
        self.pushButton_update_pattern.setText(_translate("Form", "Update Pattern"))
        self.label_milling_current.setText(_translate("Form", "Current (nA)"))
        self.label_patterns_header.setText(_translate("Form", "Patterns"))
        self.label_spacing.setText(_translate("Form", "Spacing"))
        self.pushButton_add_milling_stage.setText(_translate("Form", "Add"))
        self.checkBox_show_milling_crosshair.setText(_translate("Form", "Show Milling Crosshair"))



import napari
from PyQt5 import QtWidgets
from datetime import datetime
from pathlib import Path
from fibsem import constants
from fibsem.ui.qtdesigner_files import image_viewer
class FibsemImageViewer(image_viewer.Ui_MainWindow,QtWidgets.QMainWindow):

    def __init__(self, viewer: napari.Viewer):
        super(FibsemImageViewer, self).__init__()
        self.setupUi(self)

        self.setMinimumWidth(400)
        self.viewer = viewer
        self.viewer.window._qt_viewer.dockLayerList.setVisible(True)
        self.viewer.window._qt_viewer.dockLayerControls.setVisible(False)

        self.image = None

        # signal when image is dragged and dropped into viewer

        self.viewer.layers.selection.events.changed.connect(self.update_selected_layer)

        self.pushButton_export.clicked.connect(self.export_metadata)

    def export_metadata(self):
        from fibsem.utils import _display_metadata
        import matplotlib.pyplot as plt
        if self.image is None: 
            napari.utils.notifications.show_info("No image loaded")
            return
        
        path = self._get_save_file_ui()

        if path is None or path == "":
            napari.utils.notifications.show_info(f"No save location selected. Image not saved")
            return

        fig = _display_metadata(self.image,show=False)

        fig.savefig(path,bbox_inches='tight',pad_inches = 0.05)

        napari.utils.notifications.show_info(f"Image Saved with metadata")


    def _get_save_file_ui( msg: str = "Select a file",_filter: str = "*png",) -> Path:
   
        path, _ = QtWidgets.QFileDialog.getSaveFileName(
            msg,
            filter=_filter,
        )
        return path


    def update_selected_layer(self,event):

        from fibsem.structures import FibsemImage, FibsemImageMetadata
        from fibsem import constants


        selected_layer = self.viewer.layers.selection.active

        if selected_layer is None:
            return

        for layer in self.viewer.layers:

            layer.visible = False

        selected_layer.visible = True


        self.image = FibsemImage.load(selected_layer.source.path)

        try:
            image_settings = self.image.metadata.image_settings

            beam_type = image_settings.beam_type.name
            resolution = f'{image_settings.resolution[0]} x {image_settings.resolution[1]}'
            hfw = image_settings.hfw * constants.SI_TO_MICRO
            name = image_settings.filename

            self.label_beamType.setText(beam_type)
            self.label_resolution.setText(resolution)
            self.label_hfw.setText(f'{hfw:.2f}')
            self.label_name.setText(name)
            self.label_metadata_version.setText(self.image.metadata.version)

            timestamp = datetime.fromtimestamp(self.image.metadata.microscope_state.timestamp)
            self.label_timestamp.setText(timestamp.strftime("%Y-%m-%d %H:%M:%S"))
        except:
            pass
        
        try:
            experiment_data = self.image.metadata.experiment
            self._load_experiment_info(experiment_data)
        except:
            pass

        try:

            position = self.image.metadata.microscope_state.stage_position

            self.label_position.setText(f'x: {position.x:.2f} y: {position.y:.2f} z: {position.z:.2f} r: {position.r:.2f} t: {position.t:.2f}')
        except:
            pass

        try:
            if beam_type == "ION":

                detector_data = self.image.metadata.microscope_state.ion_detector
                beam_data = self.image.metadata.microscope_state.ion_beam
            
            else:

                detector_data = self.image.metadata.microscope_state.electron_detector
                beam_data = self.image.metadata.microscope_state.electron_beam

            self._load_detector_info(detector_data)
        except:
            pass

        try:
            self._load_beam_info(beam_data)
        except:
            pass
        
        try:
            self._load_system_info(self.image.metadata.system)
        except:
            pass

        try:
            self._load_user_info(self.image.metadata.user)
        except:
            pass


    def _load_experiment_info(self,experiment):

        if experiment is None:
            text = "No Information"
            self.label_ID.setText(text)
            self.label_method.setText(text)
            self.label_application.setText(text)
            self.label_app_version.setText("")
            self.label_fibsem_version.setText(text)

        else:
            self.label_ID.setText(experiment.id if experiment.id is not None else "")
            self.label_method.setText(experiment.method if experiment.method is not None else "")
            self.label_application.setText(experiment.application if experiment.application is not None else "")
            self.label_app_version.setText(experiment.application_version   if experiment.application_version is not None else "")
            self.label_fibsem_version.setText(experiment.fibsem_version if experiment.fibsem_version is not None else "")


    def _load_detector_info(self,detector):

        if detector is None:
            text = "No Information"
            self.label_detector_type.setText(text)
            self.label_detector_mode.setText(text)
            self.label_detector_brightness.setText(text)
            self.label_detector_contrast.setText(text)

        else:

            self.label_detector_type.setText(detector.type if detector.type is not None else "")
            self.label_detector_mode.setText(detector.mode if detector.mode is not None else "")
            self.label_detector_brightness.setText(f'{detector.brightness:.2f}' if detector.brightness is not None else "")
            self.label_detector_contrast.setText(f'{detector.contrast:.2f}' if detector.contrast is not None else "")
        
        

    def _load_beam_info(self,beam):

        if beam is None:
            text = "No Information"
            self.label_beam_current.setText(text)
            self.label_voltage.setText(text)
            self.label_working_distance.setText(text)
            self.label_dwell_time.setText(text)
            self.label_scan_rotation.setText(text)
            self.label_shift.setText(text)
            self.label_stigmation.setText(text)

        else:

            self.label_beam_current.setText(f'{beam.beam_current*constants.SI_TO_NANO:.2f} nA' if beam.beam_current is not None else "")
            self.label_voltage.setText(f'{beam.voltage*constants.SI_TO_KILO:.2f} kV' if beam.voltage is not None else "")
            self.label_working_distance.setText(f'{beam.working_distance*constants.SI_TO_MILLI:.2f} mm' if beam.working_distance is not None else "")
            self.label_dwell_time.setText(f'{beam.dwell_time*constants.SI_TO_MICRO:.2f} us' if beam.dwell_time is not None else "")
            self.label_scan_rotation.setText(f'{beam.scan_rotation:.2f}' if beam.scan_rotation is not None else "")

            shift_text = f'x: {beam.shift.x*constants.SI_TO_MICRO:.2f}um y: {beam.shift.y*constants.SI_TO_MICRO:.2f}um' if beam.shift is not None else ""
            
            self.label_shift.setText(shift_text)

            stigmation_text = f'x: {beam.stigmation.x*constants.SI_TO_MICRO:.2f}um y: {beam.stigmation.y*constants.SI_TO_MICRO:.2f}um' if beam.stigmation is not None else ''

            self.label_stigmation.setText(stigmation_text)





    def _load_system_info(self,system):

        if system is None:
            text = "No Information"
            self.label_manufacturer.setText(text)
            self.label_model.setText(text)
            self.label_serial_no.setText(text)
            self.label_software_version.setText(text)

        else:

            self.label_manufacturer.setText(system.manufacturer if system.manufacturer is not None else "")
            self.label_model.setText(system.model if system.model is not None else "")
            self.label_serial_no.setText(system.serial_number if system.serial_number is not None else "")
            self.label_software_version.setText(system.software_version if system.software_version is not None else "")

    def _load_user_info(self,user):

        if user is None:
            text = "No Information"
            self.label_username.setText(text)
            self.label_email.setText(text)
            self.label_organisation.setText(text)
            self.label_userpc.setText(text)

        else:

            self.label_username.setText(user.name if user.name is not None else "")
            self.label_email.setText(user.email if user.email is not None else "")
            self.label_organisation.setText(user.organization if user.organization is not None else "")
            self.label_userpc.setText(user.computer if user.computer is not None else "")


        





def main():
    viewer = napari.Viewer(ndisplay=2)
    image_viewer_window = FibsemImageViewer(viewer=viewer)
    viewer.window.add_dock_widget(image_viewer_window, 
                                    area="right", 
                                        add_vertical_stretch=True, 
                                        name=f"OpenFIBSEM Image Viewer")

    napari.run(max_loop_level=2)


if __name__ == "__main__":
    main()


import argparse
import napari
import napari.utils.notifications
import fibsem
from PyQt5 import QtWidgets
from fibsem import config as cfg
from fibsem.ui.qtdesigner_files import FibsemMicroscopeConfigurationWidget
from fibsem.ui.FibsemMicroscopeConfigurationWidgetBase import FibsemMicroscopeConfigurationWidgetBase


class FibsemMicroscopeConfigurationWidget(FibsemMicroscopeConfigurationWidget.Ui_MainWindow, QtWidgets.QMainWindow):
    def __init__(
        self,
        path: str = None,
        viewer: napari.Viewer = None,
        parent=None,
    ):
        super(FibsemMicroscopeConfigurationWidget, self).__init__(parent=parent)
        self.setupUi(self)
        self.setWindowTitle("Microscope Configuration")

        self.viewer = viewer

        self.configuration_widget = FibsemMicroscopeConfigurationWidgetBase(path)
        # add to layout
        self.gridLayout.addWidget(self.configuration_widget)

        self.setup_connections()

    def setup_connections(self):

        # actions
        self.actionSave_Configuration.triggered.connect(self.configuration_widget.save_configuration)
        self.actionLoad_Configuration.triggered.connect(self.configuration_widget.load_configuration_from_file)


def main():

    # parse arguments
    parser = argparse.ArgumentParser(f"Microscope Configuration UI")
    parser.add_argument("--config", type=str, default=cfg.DEFAULT_CONFIGURATION_PATH, help="Path to microscope configuration file")
    args = parser.parse_args()

    # widget viewer
    viewer = napari.Viewer(ndisplay=2)
    microscope_configuration = FibsemMicroscopeConfigurationWidget(path=args.config, viewer=viewer)
    viewer.window.add_dock_widget(
        microscope_configuration, 
        area="right", 
        add_vertical_stretch=False,
        name=f"OpenFIBSEM v{fibsem.__version__} Microscope Configuration",
    )
    napari.run()


if __name__ == "__main__":
    main()


import logging
import numpy as np

from skimage import exposure
from fibsem.structures import FibsemImage

def auto_gamma(
    image: FibsemImage,
    min_gamma: float = 0.15,
    max_gamma: float = 1.8,
    scale_factor: float = 0.01,
    gamma_threshold: int = 45,
    method: str = "autogamma",
) -> FibsemImage:
    """
    Applies automatic gamma correction to the input `FibsemImage`.

    Args:
        image (FibsemImage): The input `FibsemImage` to apply gamma correction to.
        min_gamma (float): The minimum gamma value allowed in the correction. Defaults to 0.15.
        max_gamma (float): The maximum gamma value allowed in the correction. Defaults to 1.8.
        scale_factor (float): A scaling factor to adjust the gamma correction range based on the image
            brightness. Defaults to 0.01.
        gamma_threshold (int): The maximum threshold of brightness difference from the mid-gray value
            (i.e., 128) before the gamma value is forced to 1.0. Defaults to 45.

    Returns:
        A new `FibsemImage` object containing the gamma-corrected image data, with the same metadata
        as the input image.

    Notes:
        - This function applies gamma correction to the input image using the `skimage.exposure.adjust_gamma`
          function, with the gamma value computed based on the mean intensity of the image.
        - If the difference between the mean image intensity and the mid-gray value (i.e., 128) is greater
          than the specified `gamma_threshold`, the gamma value is forced to 1.0 to avoid over-correction.
        - The `FibsemImage` object in the returned list contains the gamma-corrected image data as a
          numpy array, as well as other image metadata.
    """

    if method == "autogamma":
        std = np.std(image.data)  # unused variable?
        mean = np.mean(image.data)
        diff = mean - 255 / 2.0
        gam = np.clip(
            min_gamma, 1 + diff * scale_factor, max_gamma
        )
        if abs(diff) < gamma_threshold:
            gam = 1.0
        if image.metadata is not None:
            logging.debug(
                f"AUTO_GAMMA | {image.metadata.image_settings.beam_type} | {diff:.3f} | {gam:.3f}"
            )
        image_data = exposure.adjust_gamma(image.data, gam)

        image = FibsemImage(data=image_data, metadata=image.metadata)
    
    if method == "autoclahe":
        image = apply_clahe(image)

    return image



def apply_clahe(
    image: FibsemImage,
    which_package: str = "skimage",
    clip_limit_cv2: float = 15,
    tile_grid_size: int = 8,
    clip_limit_skimage: float = 0.02,
    kernel_size = None
) -> FibsemImage:
    """
    Applies Contrast Limited Adaptive Histogram Equalisation correction to the input `FibsemImage`.
    image is divided into small blocks called "tiles" (tileSize is 8x8 by default in OpenCV). Then each of these
    blocks are histogram equalized as usual. So in a small area, histogram would confine to a small region
    (unless there is noise). If noise is there, it will be amplified. To avoid this, contrast limiting is applied.
    If any histogram bin is above the specified contrast limit (by default 40 in OpenCV), those pixels are clipped and
    distributed uniformly to other bins before applying histogram equalization. After equalization, to remove artifacts
    in tile borders, bilinear interpolation is applied.

    In OpenCV tileGridSize (tile_grid_size) is by default 8x8, clipLimit (clip_limit_cv2) is by default 40

    In skimage kernel_size (int or array_like), is optional. It defines the shape of contextual regions used in the
    algorithm. By default, kernel_size is 1/8 of image height by 1/8 of its width.
    clip_limit float, optional. Clipping limit, normalized between 0 and 1 (higher values give more contrast).

    Args:
        image (FibsemImage): The input `FibsemImage` to apply gamma correction to.

        type (str) Either "skimage" or "OpenCV" to apply the filter from the corresponding library

        clip_limit_cv2 (float): used if which_package=="OpenCV". Defaults to 15. (by default 40 in OpenCV)
        tile_grid_size (int): used if which_package=='OpenCV'. Defaults to 8x8 pixels.

        clip_limit_skimage (float): used if which_package=="skimage". Defaults to 0.01. Clipping limit, normalised between 0 and 1 (higher values give more contrast).
        tile_grid_size (int): used if which_package=='skimage'. if None, defaults to kernel_size is 1/8 of image height by 1/8 of its width.

    Returns:
        A new `FibsemImage` object containing the clahe-enhanced image data, with the same metadata
        as the input image.

    Notes:
        - This function applies gamma correction to the input image using either the `cv2.createCLAHE` or `skimage.exposure.equalize_adapthist` function
        - The `FibsemImage` object in the returned list contains the clahe-enhanced image data as a
          numpy array, as well as other image metadata.
    """

    """
        OpenCV requires 8-bit images for CLAHE, skimage requires either 8-bit images or arrays with values between [0,1]
        Here, we convert the raw data into an 8-bit image to proceed
    """

    temp = image.data
    temp = temp / temp.max()
    temp = (temp * 2**8).astype(np.uint8)

    if which_package=='OpenCV':
        import cv2
        tile_grid_size = int(tile_grid_size)
        clahe = cv2.createCLAHE(clipLimit=clip_limit_cv2,
                                tileGridSize=(tile_grid_size,tile_grid_size))
        image_data = clahe.apply(temp)

    else: # default filter
        # nbin = 256 default, for 8-bit images
        image_data = exposure.equalize_adapthist(temp,
                                                 kernel_size=kernel_size,
                                                 clip_limit=clip_limit_skimage, nbins=256)
        import skimage
        image_data = skimage.img_as_ubyte(image_data)

    return FibsemImage(data=image_data, metadata=image.metadata)




import pandas as pd
import numpy as np

def _parse_health_monitor_data(path: str) -> pd.DataFrame:
    # PATH = "health-monitor/data2.csv"

    df = pd.read_csv(path, skiprows=5)

    # display(df)


    # header = system
    # row 0 = subsystem
    # row 1 = component
    # row 2 = parameter
    # row 3 = unit
    # row 4 = type

    # col 0 = date
    # col 1 = timestamp

    df_headers = df.iloc[0:5, 2:]
    df_data = df.iloc[6:, :]

    systems = df_headers.columns.values 
    subsystems = df_headers.iloc[0, :].values
    components = df_headers.iloc[1, :].values
    parameters = df_headers.iloc[2, :].values
    units = df_headers.iloc[3, :].values
    types = df_headers.iloc[4, :].values

    type_map = {
        "Int": np.uint8,
        "Float": np.float32,
        "String": str,
        "Boolean": bool,
        "DateTime": np.datetime64,
    }

    new_columns = ["Date", "Time"]
    new_columns_type = ["datetime64", "datetime64"]
    for subsystem, component, parameter, unit, type in zip(subsystems, components, parameters, units, types):
        new_columns.append(f"{subsystem}.{component}.{parameter} ({unit})")
        new_columns_type.append(type_map[type])

    df_data.columns = new_columns

    # replace all values of "NaN" with np.nan
    df_data = df_data.replace("NaN", np.nan)

    # drop columns with all NaN values
    # df_data = df_data.dropna(axis=0, how="all")
    # df_data = df_data.astype(dict(zip(new_columns, new_columns_type)))

    # combine date and time columns
    df_data["datetime"] = pd.to_datetime(df_data["Date"] + " " + df_data["Time"])

    # set timezone to Aus/Sydney for datetime column
    # df_data["datetime"] = df_data["datetime"].dt.tz_localize("UTC").dt.tz_convert("Australia/Sydney")

    # drop Date and Time columns
    df_data = df_data.drop(columns=["Date", "Time"])
    

    # print duplicate columns
    # drop duplicate columns
    df_data = df_data.loc[:,~df_data.columns.duplicated()]
    print(df_data.columns[df_data.columns.duplicated()])

    
    return df_data



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemSegmentationModelWidget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(456, 352)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.checkpoint_seg_button = QtWidgets.QToolButton(Form)
        self.checkpoint_seg_button.setObjectName("checkpoint_seg_button")
        self.gridLayout.addWidget(self.checkpoint_seg_button, 2, 2, 1, 1)
        self.label_model_type = QtWidgets.QLabel(Form)
        self.label_model_type.setObjectName("label_model_type")
        self.gridLayout.addWidget(self.label_model_type, 1, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 4, 0, 1, 2)
        self.label_header_model = QtWidgets.QLabel(Form)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_header_model.setFont(font)
        self.label_header_model.setObjectName("label_header_model")
        self.gridLayout.addWidget(self.label_header_model, 0, 0, 1, 2)
        self.label_checkpoint = QtWidgets.QLabel(Form)
        self.label_checkpoint.setObjectName("label_checkpoint")
        self.gridLayout.addWidget(self.label_checkpoint, 2, 0, 1, 1)
        self.comboBox_model_type = QtWidgets.QComboBox(Form)
        self.comboBox_model_type.setObjectName("comboBox_model_type")
        self.gridLayout.addWidget(self.comboBox_model_type, 1, 1, 1, 1)
        self.lineEdit_checkpoint = QtWidgets.QLineEdit(Form)
        self.lineEdit_checkpoint.setObjectName("lineEdit_checkpoint")
        self.gridLayout.addWidget(self.lineEdit_checkpoint, 2, 1, 1, 1)
        self.pushButton_load_model = QtWidgets.QPushButton(Form)
        self.pushButton_load_model.setObjectName("pushButton_load_model")
        self.gridLayout.addWidget(self.pushButton_load_model, 3, 0, 1, 2)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.checkpoint_seg_button.setText(_translate("Form", "..."))
        self.label_model_type.setText(_translate("Form", "Model"))
        self.label_header_model.setText(_translate("Form", "Segmentation Model"))
        self.label_checkpoint.setText(_translate("Form", "Checkpoint"))
        self.pushButton_load_model.setText(_translate("Form", "Load Model"))



# slice and view

import os
import logging
from pprint import pprint

import fibsem
from fibsem import acquire, milling, utils, alignment
from fibsem.structures import BeamType, ImageSettings, FibsemRectangleSettings
import numpy as np

def main():

    PROTOCOL_PATH = os.path.join(os.path.dirname(__file__), "protocol_slice_and_view.yaml")
    microscope, settings = utils.setup_session(protocol_path = PROTOCOL_PATH)

    # setup for milling
    milling.setup_milling(microscope = microscope,
        patterning_mode  = "Serial",
        mill_settings = settings.milling)
    
    pattern_settings = FibsemRectangleSettings(
        width = settings.protocol["milling"]["width"],
        height = settings.protocol["milling"]["height"],
        depth = settings.protocol["milling"]["depth"],
        scan_direction= settings.protocol["milling"]["scan_direction"],
        cleaning_cross_section= settings.protocol["milling"]["cleaning_cross_section"]
    )

    # angle correction
    microscope.set("angular_correction_tilt_correction", True)
    microscope.set("angular_correction_angle", np.deg2rad(-38))

    # update image settings
    settings.image.filename = "reference"
    settings.image.save = True
    settings.image.beam_type = BeamType.ELECTRON

    eb_image, ib_image = acquire.take_reference_images(microscope, settings.image)
    ref_eb, ref_ib = None, None

    for slice_idx in range(int(settings.protocol["steps"])):

        # slice
        logging.info("------------------------ SLICE ------------------------")
        milling_settings = settings.milling  

        patterns = milling.draw_pattern(microscope, pattern_settings)
        # estimated_milling_time = milling.estimate_milling_time_in_seconds([patterns])
        # logging.info(f"Estimated milling time: {estimated_milling_time}")
        milling.run_milling(microscope, milling_current=milling_settings.milling_current)
        milling.finish_milling(microscope, settings.system.ion.current)

        # view
        logging.info("------------------------ VIEW ------------------------")
        settings.image.filename = f"slice_{slice_idx:04d}"
        eb_image = acquire.new_image(microscope, settings.image)


        # align
        if ref_eb is not None:
            alignment.align_using_reference_images(microscope, settings, ref_eb, eb_image)
            ref_eb = eb_image

    
        # update patterns
        pattern_settings.centre_y += settings.protocol["step_size"]


if __name__ == "__main__":
    main()



import numpy as np
MICRON_TO_METRE = 1e-6
METRE_TO_MICRON = 1.0 / MICRON_TO_METRE

MILLIMETRE_TO_METRE = 1e-3
METRE_TO_MILLIMETRE = 1.0 / MILLIMETRE_TO_METRE

SI_TO_KILO = 1e-3
KILO_TO_SI =  1/ SI_TO_KILO

SI_TO_MILLI = 1e3
MILLI_TO_SI = 1 / SI_TO_MILLI

SI_TO_MICRO = 1e6
MICRO_TO_SI = 1 / SI_TO_MICRO

SI_TO_NANO = 1e9
NANO_TO_SI = 1 / SI_TO_NANO

SI_TO_PICO = 1e12
PICO_TO_SI = 1 / SI_TO_PICO

RADIANS_TO_DEGREES = 180.0 / np.pi

DEGREES_TO_RADIANS = np.pi / 180.0

TO_PERCENTAGES = 100.0

FROM_PERCENTAGES = 1.0 / TO_PERCENTAGES


# fibsem structures

import json
import os
import sys
from dataclasses import dataclass
from datetime import datetime
from enum import Enum, auto
from pathlib import Path
from fibsem.config import SUPPORTED_COORDINATE_SYSTEMS
from typing import Optional
import numpy as np
import tifffile as tff
import fibsem
from fibsem.config import METADATA_VERSION
from abc import ABC, abstractmethod, abstractstaticmethod

try:
    from tescanautomation.Common import Document

    TESCAN = True
except:
    TESCAN = False

try:
    sys.path.append("C:\Program Files\Thermo Scientific AutoScript")
    sys.path.append(
        "C:\Program Files\Enthought\Python\envs\AutoScript\Lib\site-packages"
    )
    sys.path.append("C:\Program Files\Python36\envs\AutoScript")
    sys.path.append("C:\Program Files\Python36\envs\AutoScript\Lib\site-packages")
    from autoscript_sdb_microscope_client.structures import (
        AdornedImage,
        ManipulatorPosition,
        Rectangle,
        StagePosition,
    )

    THERMO = True
except:
    THERMO = False



# @patrickcleeve: dataclasses.asdict -> :(


# TODO: overload constructors instead of from_dict...
@dataclass
class Point:
    x: float = 0.0
    y: float = 0.0
    name: Optional[str] = None

    def to_dict(self) -> dict:
        return {"x": self.x, "y": self.y}

    @staticmethod
    def from_dict(d: dict) -> "Point":
        x = float(d["x"])
        y = float(d["y"])
        return Point(x, y)

    def to_list(self) -> list:
        return [self.x, self.y]

    @staticmethod
    def from_list(l: list) -> "Point":
        x = float(l[0])
        y = float(l[1])
        return Point(x, y)

    def __add__(self, other) -> "Point":
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other) -> "Point":
        return Point(self.x - other.x, self.y - other.y)

    def __len__(self) -> int:
        return 2

    def __getitem__(self, key: int) -> float:
        if key == 0:
            return self.x
        elif key == 1:
            return self.y
        else:
            raise IndexError("Index out of range")

    def _to_metres(self, pixel_size: float) -> "Point":
        return Point(self.x * pixel_size, self.y * pixel_size)

    def _to_pixels(self, pixel_size: float) -> "Point":
        return Point(self.x / pixel_size, self.y / pixel_size)

    def distance(self, other: "Point") -> "Point":
        """Calculate the distance between two points. (other - self)"""
        return Point(x=(other.x - self.x), y=(other.y - self.y))

    def euclidean(self, other: "Point") -> float:
        """Calculate the euclidean distance between two points."""
        return np.linalg.norm(self.distance(other).to_list())


# TODO: convert these to match autoscript...
class BeamType(Enum):
    """Enumerator Class for Beam Type
    1: Electron Beam
    2: Ion Beam

    """

    ELECTRON = 1  # Electron
    ION = 2  # Ion
    # CCD_CAM = 3
    # NavCam = 4 # see enumerations/ImagingDevice


class MovementMode(Enum):
    Stable = 1
    Vertical = 2
    # Needle = 3


class ImagingState:
    IDLE = 0
    ACQUIRING = 1
    STOPPING = 2
    PAUSED = 3
    ERROR = 4

class PatterningState:
    IDLE = 0
    PATTERNING = 1
    STOPPING = 2
    PAUSED = 3
    ERROR = 4

class ManipulatorState(Enum):
    RETRACTED = 0
    INSERTED = 1
    MOVING = 2


@dataclass
class FibsemStagePosition:
    """Data class for storing stage position data.

    Attributes:
        x (float): The X position of the stage in meters.
        y (float): The Y position of the stage in meters.
        z (float): The Z position of the stage in meters.
        r (float): The Rotation of the stage in radians.
        t (float): The Tilt of the stage in radians.
        coordinate_system (str): The coordinate system used for the stage position.

    Methods:
        to_dict(): Convert the stage position object to a dictionary.
        from_dict(data: dict): Create a new stage position object from a dictionary.
        to_autoscript_position(stage_tilt: float = 0.0) -> StagePosition: Convert the stage position to a StagePosition object that is compatible with Autoscript.
        from_autoscript_position(position: StagePosition, stage_tilt: float = 0.0) -> None: Create a new FibsemStagePosition object from a StagePosition object that is compatible with Autoscript.
        to_tescan_position(stage_tilt: float = 0.0): Convert the stage position to a format that is compatible with Tescan.
        from_tescan_position(): Create a new FibsemStagePosition object from a Tescan-compatible stage position.
    """

    name: str = None
    x: float = None
    y: float = None
    z: float = None
    r: float = None
    t: float = None
    coordinate_system: str = None

    def to_dict(self) -> dict:
        position_dict = {}

        position_dict["name"] = self.name if self.name is not None else None
        position_dict["x"] = float(self.x) if self.x is not None else None
        position_dict["y"] = float(self.y) if self.y is not None else None
        position_dict["z"] = float(self.z) if self.z is not None else None
        position_dict["r"] = float(self.r) if self.r is not None else None
        position_dict["t"] = float(self.t) if self.t is not None else None
        position_dict["coordinate_system"] = self.coordinate_system

        return position_dict

    @classmethod
    def from_dict(cls, data: dict) -> "FibsemStagePosition":
        items = ["x", "y", "z", "r", "t"]

        for item in items:
            value = data[item]

            assert isinstance(value, float) or isinstance(value, int) or value is None

        return cls(
            name=data.get("name", None),
            x=data["x"],
            y=data["y"],
            z=data["z"],
            r=data["r"],
            t=data["t"],
            coordinate_system=data["coordinate_system"],
        )

    if THERMO:

        def to_autoscript_position(self, stage_tilt: float = 0.0) -> StagePosition:
            return StagePosition(
                x=self.x,
                y=self.y,  # / np.cos(stage_tilt),
                z=self.z,  # / np.cos(stage_tilt),
                r=self.r,
                t=self.t,
                coordinate_system=self.coordinate_system,
            )

        @classmethod
        def from_autoscript_position(
            cls, position: StagePosition, stage_tilt: float = 0.0
        ) -> None:
            return cls(
                x=position.x,
                y=position.y,  # * np.cos(stage_tilt),
                z=position.z,  # * np.cos(stage_tilt),
                r=position.r,
                t=position.t,
                coordinate_system=position.coordinate_system.upper(),
            )

    if TESCAN:

        def to_tescan_position(self, stage_tilt: float = 0.0):
            self.y = self.y  # / np.cos(stage_tilt),

        @classmethod
        def from_tescan_position(self, stage_tilt: float = 0.0):
            self.y = self.y  # * np.cos(stage_tilt)

    def __add__(self, other: "FibsemStagePosition") -> "FibsemStagePosition":
        return FibsemStagePosition(
            x=self.x + other.x if other.x is not None else self.x,
            y=self.y + other.y if other.y is not None else self.y,
            z=self.z + other.z if other.z is not None else self.z,
            r=self.r + other.r if other.r is not None else self.r,
            t=self.t + other.t if other.t is not None else self.t,
            coordinate_system=self.coordinate_system,
        )

    def __sub__(self, other: "FibsemStagePosition") -> "FibsemStagePosition":
        return FibsemStagePosition(
            x=self.x - other.x,
            y=self.y - other.y,
            z=self.z - other.z,
            r=self.r - other.r,
            t=self.t - other.t,
            coordinate_system=self.coordinate_system,
        )

    def _scale_repr(self, scale: float, precision: int = 2):
        return f"x:{self.x*scale:.{precision}f}, y:{self.y*scale:.{precision}f}, z:{self.z*scale:.{precision}f}"



@dataclass
class FibsemManipulatorPosition:
    """Data class for storing manipulator position data.

    Attributes:
        x (float): The X position of the manipulator in meters.
        y (float): The Y position of the manipulator in meters.
        z (float): The Z position of the manipulator in meters.
        r (float): The Rotation of the manipulator in radians.
        t (float): The Tilt of the manipulator in radians.
        coordinate_system (str): The coordinate system used for the manipulator position.

    Methods:
        to_dict(): Convert the manipulator position object to a dictionary.
        from_dict(data: dict): Create a new manipulator position object from a dictionary.
        to_autoscript_position() -> ManipulatorPosition: Convert the manipulator position to a ManipulatorPosition object that is compatible with Autoscript.
        from_autoscript_position(position: ManipulatorPosition) -> None: Create a new FibsemManipulatorPosition object from a ManipulatorPosition object that is compatible with Autoscript.
        to_tescan_position(): Convert the manipulator position to a format that is compatible with Tescan.
        from_tescan_position(): Create a new FibsemManipulatorPosition object from a Tescan-compatible manipulator position.
    """

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    r: float = 0.0
    t: float = 0.0
    coordinate_system: str = "RAW"

    def __post_init__(self):
        assert (
            isinstance(self.coordinate_system, str) or self.coordinate_system is None
        ), f"unsupported type {type(self.coordinate_system)} for coorindate system"
        assert (
            self.coordinate_system in SUPPORTED_COORDINATE_SYSTEMS
            or self.coordinate_system is None
        ), f"coordinate system value {self.coordinate_system} is unsupported or invalid syntax. Must be RAW or SPECIMEN"

    def to_dict(self) -> dict:
        position_dict = {}
        position_dict["x"] = self.x
        position_dict["y"] = self.y
        position_dict["z"] = self.z
        position_dict["r"] = self.r
        position_dict["t"] = self.t
        position_dict["coordinate_system"] = self.coordinate_system.upper()

        return position_dict

    @classmethod
    def from_dict(cls, data: dict) -> "FibsemManipulatorPosition":
        items = ["x", "y", "z", "r", "t"]

        for item in items:
            value = data[item]

            assert isinstance(value, float) or isinstance(value, int) or value is None

        return cls(
            x=data["x"],
            y=data["y"],
            z=data["z"],
            r=data["r"],
            t=data["t"],
            coordinate_system=data["coordinate_system"],
        )

    if THERMO:

        def to_autoscript_position(self) -> ManipulatorPosition:
            if self.coordinate_system == "RAW":
                coordinate_system = "Raw"
            elif self.coordinate_system == "STAGE":
                coordinate_system = "Stage"
            return ManipulatorPosition(
                x=self.x,
                y=self.y,
                z=self.z,
                r=None,  # TODO figure this out, do we need it for real micrscope or just simulator ?
                # r=None,
                coordinate_system=coordinate_system,
            )

        @classmethod
        def from_autoscript_position(cls, position: ManipulatorPosition) -> None:
            return cls(
                x=position.x,
                y=position.y,
                z=position.z,
                coordinate_system=position.coordinate_system.upper(),
            )

    if TESCAN:

        def to_tescan_position(self):
            pass

        @classmethod
        def from_tescan_position(self):
            pass

    def __add__(
        self, other: "FibsemManipulatorPosition"
    ) -> "FibsemManipulatorPosition":
        return FibsemManipulatorPosition(
            self.x + other.x,
            self.y + other.y,
            self.z + other.z,
            self.r + other.r,
            self.t + other.t,
            self.coordinate_system,
        )


@dataclass
class FibsemRectangle:
    """Universal Rectangle class used for ReducedArea"""

    left: float = 0.0
    top: float = 0.0
    width: float = 0.0
    height: float = 0.0

    def __post_init__(self):
        assert isinstance(self.left, float) or isinstance(
            self.left, int
        ), f"type {type(self.left)} is unsupported for left, must be int or floar"
        assert isinstance(self.top, float) or isinstance(
            self.top, int
        ), f"type {type(self.top)} is unsupported for top, must be int or floar"
        assert isinstance(self.width, float) or isinstance(
            self.width, int
        ), f"type {type(self.width)} is unsupported for width, must be int or floar"
        assert isinstance(self.height, float) or isinstance(
            self.height, int
        ), f"type {type(self.height)} is unsupported for height, must be int or floar"

    def from_dict(settings: dict) -> "FibsemRectangle":
        if settings is None:
            return None
        points = ["left", "top", "width", "height"]

        for point in points:
            value = settings[point]

            assert isinstance(value, float) or isinstance(value, int) or value is None

        return FibsemRectangle(
            left=settings["left"],
            top=settings["top"],
            width=settings["width"],
            height=settings["height"],
        )

    def to_dict(self) -> dict:
        return {
            "left": float(self.left),
            "top": float(self.top),
            "width": float(self.width),
            "height": float(self.height),
        }

    if THERMO:

        def __to_FEI__(self) -> Rectangle:
            return Rectangle(self.left, self.top, self.width, self.height)

        @classmethod
        def __from_FEI__(cls, rect: Rectangle) -> "FibsemRectangle":
            return cls(rect.left, rect.top, rect.width, rect.height)


@dataclass
class ImageSettings:
    """A data class representing the settings for an image acquisition.

    Attributes:
        resolution (list of int): The resolution of the acquired image in pixels, [x, y].
        dwell_time (float): The time spent per pixel during image acquisition, in seconds.
        hfw (float): The horizontal field width of the acquired image, in microns.
        autocontrast (bool): Whether or not to apply automatic contrast enhancement to the acquired image.
        beam_type (BeamType): The type of beam to use for image acquisition.
        save (bool): Whether or not to save the acquired image to disk.
        filename (str): The filename to use when saving the acquired image.
        autogamma (bool): Whether or not to apply gamma correction to the acquired image.
        path (Path): The path to the directory where the acquired image should be saved.
        reduced_area (FibsemRectangle): The rectangular region of interest within the acquired image, if any.

    Methods:
        from_dict(settings: dict) -> ImageSettings:
            Converts a dictionary of image settings to an ImageSettings object.
        to_dict() -> dict:
            Converts the ImageSettings object to a dictionary of image settings.
    """

    resolution: list = None
    dwell_time: float = None
    hfw: float = None
    autocontrast: bool = None
    beam_type: BeamType = None
    save: bool = None
    filename: str = None
    autogamma: bool = None
    path: Path = None
    reduced_area: FibsemRectangle = None
    line_integration: int = None  # (int32) 2 - 255
    scan_interlacing: int = None  # (int32) 2 - 8
    frame_integration: int = None  # (int32) 2 - 512
    drift_correction: bool = False  # (bool) # requires frame_integration > 1

    def __post_init__(self):
        assert (
            isinstance(self.resolution, (list, tuple)) or self.resolution is None
        ), f"resolution must be a list, currently is {type(self.resolution)}"
        assert (
            isinstance(self.dwell_time, float) or self.dwell_time is None
        ), f"dwell time must be of type float, currently is {type(self.dwell_time)}"
        assert (
            isinstance(self.hfw, float) or isinstance(self.hfw, int) or self.hfw is None
        ), f"hfw must be int or float, currently is {type(self.hfw)}"
        assert (
            isinstance(self.autocontrast, bool) or self.autocontrast is None
        ), f"autocontrast setting must be bool, currently is {type(self.autocontrast)}"
        assert (
            isinstance(self.beam_type, BeamType) or self.beam_type is None
        ), f"beam type must be a BeamType object, currently is {type(self.beam_type)}"
        assert (
            isinstance(self.save, bool) or self.save is None
        ), f"save option must be a bool, currently is {type(self.save)}"
        assert (
            isinstance(self.filename, str) or self.filename is None
        ), f"filename must b str, currently is {type(self.filename)}"
        assert (
            isinstance(self.autogamma, bool) or self.autogamma is None
        ), f"gamma enabled setting must be bool, currently is {type(self.autogamma)}"
        assert (
            isinstance(self.path, (Path, str)) or self.path is None
        ), f"save path must be Path or str, currently is {type(self.path)}"
        assert (
            isinstance(self.reduced_area, FibsemRectangle) or self.reduced_area is None
        ), f"reduced area must be a fibsemRectangle object, currently is {type(self.reduced_area)}"

    @staticmethod
    def from_dict(settings: dict) -> "ImageSettings":
        if "reduced_area" in settings and settings["reduced_area"] is not None:
            reduced_area = FibsemRectangle.from_dict(settings["reduced_area"])
        else:
            reduced_area = None

        image_settings = ImageSettings(
            resolution=settings.get("resolution", (1536, 1024)),
            dwell_time=settings.get("dwell_time", 1.0e-6),
            hfw=settings.get("hfw", 150e-6),
            autocontrast=settings.get("autocontrast", False),
            beam_type=BeamType[settings.get("beam_type", "Electron").upper()],
            autogamma=settings.get("autogamma", False),
            save=settings.get("save", False),
            path=settings.get("path", os.getcwd()),
            filename=settings.get("filename", "default_image"),
            reduced_area=reduced_area,
            line_integration=settings.get("line_integration", None),
            scan_interlacing=settings.get("scan_interlacing", None),
            frame_integration=settings.get("frame_integration", None),
            drift_correction=settings.get("drift_correction", False),
        )

        return image_settings

    def to_dict(self) -> dict:
        settings_dict = {
            "beam_type": self.beam_type.name if self.beam_type is not None else None,
            "resolution": self.resolution if self.resolution is not None else None,
            "dwell_time": self.dwell_time if self.dwell_time is not None else None,
            "hfw": self.hfw if self.hfw is not None else None,
            "autocontrast": self.autocontrast
            if self.autocontrast is not None
            else None,
            "autogamma": self.autogamma
            if self.autogamma is not None
            else None,
            "save": self.save if self.save is not None else None,
            "path": str(self.path) if self.path is not None else None,
            "filename": self.filename if self.filename is not None else None,
            "reduced_area": {
                "left": self.reduced_area.left,
                "top": self.reduced_area.top,
                "width": self.reduced_area.width,
                "height": self.reduced_area.height,
            }
            if self.reduced_area is not None
            else None,
            "line_integration": self.line_integration,
            "scan_interlacing": self.scan_interlacing,
            "frame_integration": self.frame_integration,
            "drift_correction": self.drift_correction,
        }

        return settings_dict

    @staticmethod
    def fromFibsemImage(image: "FibsemImage") -> "ImageSettings":
        """Returns the image settings for a FibsemImage object.

        Args:
            image (FibsemImage): The FibsemImage object to get the image settings from.

        Returns:
            ImageSettings: The image settings for the given FibsemImage object.
        """
        from fibsem import utils
        from copy import deepcopy

        image_settings = deepcopy(image.metadata.image_settings)
        image_settings.filename = utils.current_timestamp()
        image_settings.save = True

        return image_settings


@dataclass
class BeamSettings:
    """
    Dataclass representing the beam settings for an imaging session.

    Attributes:
        beam_type (BeamType): The type of beam to use for imaging.
        working_distance (float): The working distance for the microscope, in meters.
        beam_current (float): The beam current for the microscope, in amps.
        hfw (float): The horizontal field width for the microscope, in meters.
        resolution (list): The desired resolution for the image.
        dwell_time (float): The dwell time for the microscope.
        stigmation (Point): The point for stigmation correction.
        shift (Point): The point for shift correction.

    Methods:
        to_dict(): Returns a dictionary representation of the object.
        from_dict(state_dict: dict) -> BeamSettings: Returns a new BeamSettings object created from a dictionary.

    """

    beam_type: BeamType
    working_distance: float = None
    beam_current: float = None
    voltage: float = None
    hfw: float = None
    resolution: list = None
    dwell_time: float = None
    stigmation: Point = None
    shift: Point = None
    scan_rotation: float = None

    def __post_init__(self):
        assert (
            self.beam_type in [BeamType.ELECTRON, BeamType.ION]
            or self.beam_type is None
        ), f"beam_type must be instance of BeamType, currently {type(self.beam_type)}"
        assert (
            isinstance(self.working_distance, (float, int))
            or self.working_distance is None
        ), f"Working distance must be float or int, currently is {type(self.working_distance)}"
        assert (
            isinstance(self.beam_current, (float, int)) or self.beam_current is None
        ), f"beam current must be float or int, currently is {type(self.beam_current)}"
        assert (
            isinstance(self.voltage, (float, int)) or self.voltage is None
        ), f"voltage must be float or int, currently is {type(self.voltage)}"
        assert (
            isinstance(self.hfw, (float, int)) or self.hfw is None
        ), f"horizontal field width (HFW) must be float or int, currently is {type(self.hfw)}"
        assert (
            isinstance(self.resolution, list) or self.resolution is None
        ), f"resolution must be a list, currently is {type(self.resolution)}"
        assert (
            isinstance(self.dwell_time, (float, int)) or self.dwell_time is None
        ), f"dwell_time must be float or int, currently is {type(self.dwell_time)}"
        assert (
            isinstance(self.stigmation, Point) or self.stigmation is None
        ), f"stigmation must be a Point instance, currently is {type(self.stigmation)}"
        assert (
            isinstance(self.shift, Point) or self.shift is None
        ), f"shift must be a Point instance, currently is {type(self.shift)}"


    def to_dict(self) -> dict:
        state_dict = {
            "beam_type": self.beam_type.name,
            "working_distance": self.working_distance,
            "beam_current": self.beam_current,
            "voltage": self.voltage,
            "hfw": self.hfw,
            "resolution": self.resolution,
            "dwell_time": self.dwell_time,
            "stigmation": self.stigmation.to_dict()
            if self.stigmation is not None
            else None,
            "shift": self.shift.to_dict() if self.shift is not None else None,
            "scan_rotation": self.scan_rotation,
        }

        return state_dict

    @staticmethod
    def from_dict(state_dict: dict) -> "BeamSettings":
        if "stigmation" in state_dict and state_dict["stigmation"] is not None:
            stigmation = Point.from_dict(state_dict["stigmation"])
        else:
            stigmation = Point()
        if "shift" in state_dict and state_dict["shift"] is not None:
            shift = Point.from_dict(state_dict["shift"])
        else:
            shift = Point()
        
        wd = state_dict.get("working_distance", state_dict.get("eucentric_height", None))
        current = state_dict.get("beam_current", state_dict.get("current", None))

        beam_settings = BeamSettings(
            beam_type=BeamType[state_dict["beam_type"].upper()],
            working_distance=wd,
            beam_current=current,
            voltage=state_dict["voltage"],
            hfw=state_dict["hfw"],
            resolution=state_dict["resolution"],
            dwell_time=state_dict["dwell_time"],
            stigmation=stigmation,
            shift=shift,
            scan_rotation=state_dict.get("scan_rotation", 0.0),
        )

        return beam_settings


@dataclass
class FibsemDetectorSettings:
    type: str = None
    mode: str = None
    brightness: float = 0.5
    contrast: float = 0.5

    def __post_init__(self):
        assert (
            isinstance(self.type, str) or self.type is None
        ), f"type must be input as str, currently is {type(self.type)}"
        assert (
            isinstance(self.mode, str) or self.mode is None
        ), f"mode must be input as str, currently is {type(self.mode)}"
        assert (
            isinstance(self.brightness, (float, int)) or self.brightness is None
        ), f"brightness must be int or float value, currently is {type(self.brightness)}"
        assert (
            isinstance(self.contrast, (float, int)) or self.contrast is None
        ), f"contrast must be int or float value, currently is {type(self.contrast)}"

    if TESCAN:

        def to_tescan(self):
            """Converts to tescan format."""
            tescan_brightness = self.brightness * 100
            tescan_contrast = self.contrast * 100
            return tescan_brightness, tescan_contrast

    def to_dict(self) -> dict:
        """Converts to a dictionary."""
        return {
            "type": self.type,
            "mode": self.mode,
            "brightness": self.brightness,
            "contrast": self.contrast,
        }

    @staticmethod
    def from_dict(settings: dict) -> "FibsemDetectorSettings":
        """Converts from a dictionary."""
        return FibsemDetectorSettings(
            type=settings.get("type", "Unknown"),
            mode=settings.get("mode", "Unknown"),
            brightness=settings.get("brightness", 0.0),
            contrast=settings.get("contrast", 0.0),
        )


@dataclass
class MicroscopeState:

    """Data Class representing the state of a microscope with various parameters.

    Attributes:

        timestamp (float): A float representing the timestamp at which the state of the microscope was recorded. Defaults to the timestamp of the current datetime.
        stage_position (FibsemStagePosition): An instance of FibsemStagePosition representing the current absolute position of the stage. Defaults to an empty instance of FibsemStagePosition.
        electron_beam (BeamSettings): An instance of BeamSettings representing the electron beam settings. Defaults to an instance of BeamSettings with beam_type set to BeamType.ELECTRON.
        ion_beam (BeamSettings): An instance of BeamSettings representing the ion beam settings. Defaults to an instance of BeamSettings with beam_type set to BeamType.ION.

    Methods:

        to_dict(self) -> dict: Converts the current state of the Microscope to a dictionary and returns it.
        from_dict(state_dict: dict) -> "MicroscopeState": Returns a new instance of MicroscopeState with attributes created from the passed dictionary.
    """

    timestamp: float = datetime.timestamp(datetime.now())
    stage_position: FibsemStagePosition = FibsemStagePosition()
    electron_beam: BeamSettings = BeamSettings(beam_type=BeamType.ELECTRON)
    ion_beam: BeamSettings = BeamSettings(beam_type=BeamType.ION)
    electron_detector: FibsemDetectorSettings = FibsemDetectorSettings()
    ion_detector: FibsemDetectorSettings = FibsemDetectorSettings()

    def __post_init__(self):
        assert (
            isinstance(self.stage_position, FibsemStagePosition)
            or self.stage_position is None
        ), f"absolute position must be of type FibsemStagePosition, currently is {type(self.stage_position)}"
        assert (
            isinstance(self.electron_beam, BeamSettings) or self.electron_beam is None
        ), f"electron_beam must be of type BeamSettings, currently is {type(self.electron_beam)}"
        assert (
            isinstance(self.ion_beam, BeamSettings) or self.ion_beam is None
        ), f"ion_beam must be of type BeamSettings, currently us {type(self.ion_beam)}"
        assert (
            isinstance(self.electron_detector, FibsemDetectorSettings)
            or self.electron_detector is None
        ), f"electron_detector must be of type FibsemDetectorSettings, currently is {type(self.electron_detector)}"
        assert (
            isinstance(self.ion_detector, FibsemDetectorSettings)
            or self.ion_detector is None
        ), f"ion_detector must be of type FibsemDetectorSettings, currently is {type(self.ion_detector)}"

    def to_dict(self) -> dict:
        state_dict = {
            "timestamp": self.timestamp,
            "stage_position": self.stage_position.to_dict()
            if self.stage_position is not None
            else None,
            "electron_beam": self.electron_beam.to_dict()
            if self.electron_beam is not None
            else None,
            "ion_beam": self.ion_beam.to_dict()
            if self.ion_beam is not None
            else None,
            "electron_detector": self.electron_detector.to_dict()
            if self.electron_detector is not None
            else None,
            "ion_detector": self.ion_detector.to_dict()
            if self.ion_detector is not None
            else None,
        }

        return state_dict

    @staticmethod
    def from_dict(state_dict: dict) -> "MicroscopeState":
        
        # beam, and detector settings are now optional
        electron_beam, electron_detector = None, None
        ion_beam, ion_detector = None, None

        if state_dict.get("electron_beam", None) is not None:
            electron_beam = BeamSettings.from_dict(state_dict["electron_beam"])
        if state_dict.get("ion_beam", None) is not None:
            ion_beam = BeamSettings.from_dict(state_dict["ion_beam"])
        if state_dict.get("electron_detector", None) is not None:
            electron_detector = FibsemDetectorSettings.from_dict(state_dict["electron_detector"])
        if state_dict.get("ion_detector", None) is not None:
            ion_detector = FibsemDetectorSettings.from_dict(state_dict["ion_detector"])

        microscope_state = MicroscopeState(
            timestamp=state_dict["timestamp"],
            stage_position=FibsemStagePosition.from_dict(
                state_dict["stage_position"]
            ),
            electron_beam=electron_beam,
            ion_beam=ion_beam,
            electron_detector=electron_detector,
            ion_detector=ion_detector,
        )

        return microscope_state



########### Base Pattern Settings
@dataclass
class FibsemPatternSettings(ABC):

    @abstractmethod
    def to_dict(self) -> dict:
        pass
    
    @staticmethod
    def from_dict(self, data: dict) -> "FibsemPatternSettings":
        pass


class CrossSectionPattern(Enum):
    Rectangle  = auto()
    RegularCrossSection = auto()
    CleaningCrossSection = auto()

@dataclass
class FibsemRectangleSettings(FibsemPatternSettings):
    width: float
    height: float
    depth: float
    centre_x: float
    centre_y: float
    rotation: float = 0
    cleaning_cross_section: bool = False
    scan_direction: str = "TopToBottom"
    cross_section: CrossSectionPattern = CrossSectionPattern.Rectangle
    passes: int = 0

    def to_dict(self) -> dict:
        return {
            "width": self.width,
            "height": self.height,
            "depth": self.depth,
            "rotation": self.rotation,
            "centre_x": self.centre_x,
            "centre_y": self.centre_y,
            "cleaning_cross_section": self.cleaning_cross_section,
            "scan_direction": self.scan_direction,
            "cross_section": self.cross_section.name,
            "passes": self.passes,
        }

    @staticmethod
    def from_dict(data: dict) -> "FibsemRectangleSettings":
        return FibsemRectangleSettings(
            width=data["width"],
            height=data["height"],
            depth=data["depth"],
            centre_x=data["centre_x"],
            centre_y=data["centre_y"],
            cleaning_cross_section=data.get("cleaning_cross_section", False),
            rotation=data.get("rotation", 0),
            scan_direction=data.get("scan_direction", "TopToBottom"),
            cross_section=CrossSectionPattern[data.get("cross_section", "Rectangle")],
            passes=data.get("passes", 0),
        )

@dataclass
class FibsemLineSettings(FibsemPatternSettings):
    start_x: float
    end_x: float
    start_y: float
    end_y: float
    depth: float

    def to_dict(self) -> dict:
        return {
            "start_x": self.start_x,
            "end_x": self.end_x,
            "start_y": self.start_y,
            "end_y": self.end_y,
            "depth": self.depth,
        }

    @staticmethod
    def from_dict(data: dict) -> "FibsemLineSettings":
        return FibsemLineSettings(
            start_x=data["start_x"],
            end_x=data["end_x"],
            start_y=data["start_y"],
            end_y=data["end_y"],
            depth=data["depth"],
        )

@dataclass
class FibsemCircleSettings(FibsemPatternSettings):
    radius: float
    depth: float
    centre_x: float
    centre_y: float
    thickness: float = 0
    start_angle: float = 0.0
    end_angle: float = 360.0
    rotation: float = 0.0           # annulus -> thickness !=0

    def to_dict(self) -> dict:
        return {
            "radius": self.radius,
            "depth": self.depth,
            "centre_x": self.centre_x,
            "centre_y": self.centre_y,
            "start_angle": self.start_angle,
            "end_angle": self.end_angle,
            "rotation": self.rotation,
            "thickness": self.thickness,
        }

    @staticmethod
    def from_dict(data: dict) -> "FibsemCircleSettings":
        return FibsemCircleSettings(
            radius=data["radius"],
            depth=data["depth"],
            centre_x=data["centre_x"],
            centre_y=data["centre_y"],
            start_angle=data.get("start_angle", 0),
            end_angle=data.get("end_angle", 360),
            rotation=data.get("rotation", 0),
            thickness=data.get("thickness", 0),
        )


@dataclass
class FibsemBitmapSettings(FibsemPatternSettings):
    width: float
    height: float
    depth: float
    rotation: float
    centre_x: float
    centre_y: float
    path: str = None

    def to_dict(self) -> dict:
        return {
            "width": self.width,
            "height": self.height,
            "depth": self.depth,
            "rotation": self.rotation,
            "centre_x": self.centre_x,
            "centre_y": self.centre_y,
            "path": self.path,
        }

    @staticmethod
    def from_dict(data: dict) -> "FibsemBitmapSettings":
        return FibsemBitmapSettings(
            width=data["width"],
            height=data["height"],
            depth=data["depth"],
            rotation=data["rotation"],
            centre_x=data["centre_x"],
            centre_y=data["centre_y"],
            path=data["path"],
        )


@dataclass
class FibsemMillingSettings:
    """
    This class is used to store and retrieve settings for FIBSEM milling.

    Attributes:
    milling_current (float): The current used in the FIBSEM milling process. Default value is 20.0e-12 A.
    spot_size (float): The size of the beam spot used in the FIBSEM milling process. Default value is 5.0e-8 m.
    rate (float): The milling rate of the FIBSEM process. Default value is 3.0e-3 m^3/A/s.
    dwell_time (float): The dwell time of the beam at each point during the FIBSEM milling process. Default value is 1.0e-6 s.
    hfw (float): The high voltage field width used in the FIBSEM milling process. Default value is 150e-6 m.

    Methods:
    to_dict(): Converts the object attributes into a dictionary.
    from_dict(settings: dict) -> "FibsemMillingSettings": Creates a FibsemMillingSettings object from a dictionary of settings.
    """

    milling_current: float = 20.0e-12
    spot_size: float = 5.0e-8
    rate: float = 3.0e-11  # m3/A/s
    dwell_time: float = 1.0e-6  # s
    hfw: float = 150e-6
    patterning_mode: str = "Serial"
    application_file: str = "Si"
    preset: str = "30 keV; UHR imaging"
    spacing: float = 1.0
    milling_voltage: float = 30e3

    def __post_init__(self):
        assert isinstance(
            self.milling_current, (float, int)
        ), f"invalid type for milling_current, must be int or float, currently {type(self.milling_current)}"
        assert isinstance(
            self.spot_size, (float, int)
        ), f"invalid type for spot_size, must be int or float, currently {type(self.spot_size)}"
        assert isinstance(
            self.rate, (float, int)
        ), f"invalid type for rate, must be int or float, currently {type(self.rate)}"
        assert isinstance(
            self.dwell_time, (float, int)
        ), f"invalid type for dwell_time, must be int or float, currently {type(self.dwell_time)}"
        assert isinstance(
            self.hfw, (float, int)
        ), f"invalid type for hfw, must be int or float, currently {type(self.hfw)}"
        assert isinstance(
            self.patterning_mode, str
        ), f"invalid type for value for patterning_mode, must be str, currently {type(self.patterning_mode)}"
        assert isinstance(
            self.application_file, (str)
        ), f"invalid type for value for application_file, must be str, currently {type(self.application_file)}"
        assert isinstance(
            self.spacing, (float, int)
        ), f"invalid type for value for spacing, must be int or float, currently {type(self.spacing)}"
        # assert isinstance(self.preset,(str)), f"invalid type for value for preset, must be str, currently {type(self.preset)}"

    def to_dict(self) -> dict:
        settings_dict = {
            "milling_current": self.milling_current,
            "spot_size": self.spot_size,
            "rate": self.rate,
            "dwell_time": self.dwell_time,
            "hfw": self.hfw,
            "patterning_mode": self.patterning_mode,
            "application_file": self.application_file,
            "preset": self.preset,
            "spacing": self.spacing,
            "milling_voltage": self.milling_voltage,
        }

        return settings_dict

    @staticmethod
    def from_dict(settings: dict) -> "FibsemMillingSettings":
        milling_settings = FibsemMillingSettings(
            milling_current=settings.get("milling_current", 20.0e-12),
            spot_size=settings.get("spot_size", 5.0e-8),
            rate=settings.get("rate", 3.0e-11),
            dwell_time=settings.get("dwell_time", 1.0e-6),
            hfw=settings.get("hfw", 150e-6),
            patterning_mode=settings.get("patterning_mode", "Serial"),
            application_file=settings.get("application_file", "Si"),
            preset=settings.get("preset", "30 keV; 1nA"),
            spacing=settings.get("spacing", 1.0),
            milling_voltage=settings.get("milling_voltage", 30e3),
        )

        return milling_settings

# migrate to this
# v3

# TODO: start here
@dataclass
class StageSystemSettings:
    rotation_reference: float
    rotation_180: float
    shuttle_pre_tilt: float
    manipulator_height_limit: float
    enabled: bool = True
    rotation: bool = True
    tilt: bool  = True

    def to_dict(self):
        return {
            "rotation_reference": self.rotation_reference,
            "rotation_180": self.rotation_180,
            "shuttle_pre_tilt": self.shuttle_pre_tilt,
            "manipulator_height_limit": self.manipulator_height_limit,
            "enabled": self.enabled,
            "rotation": self.rotation,
            "tilt": self.tilt,
        }
    
    @staticmethod
    def from_dict(settings: dict):
        return StageSystemSettings(
            rotation_reference=settings["rotation_reference"],
            rotation_180=settings["rotation_180"],
            shuttle_pre_tilt=settings["shuttle_pre_tilt"],
            manipulator_height_limit=settings["manipulator_height_limit"],
            enabled=settings.get("enabled", True),
            rotation=settings.get("rotation", True),
            tilt=settings.get("tilt", True),
        )


@dataclass
class BeamSystemSettings:
    beam_type: BeamType
    enabled: bool
    beam: BeamSettings
    detector: FibsemDetectorSettings
    eucentric_height: float
    column_tilt: float
    plasma: bool = False
    plasma_gas: str = None

    def to_dict(self):
        ddict = {
            "beam_type": self.beam_type.value,
            "enabled": self.enabled,
            "eucentric_height": self.eucentric_height,
            "column_tilt": self.column_tilt,
            "plasma": self.plasma,
            "plasma_gas": self.plasma_gas,
        }
        ddict.update(self.beam.to_dict())
        ddict.update(self.detector.to_dict())
        
        # rename keys to match config
        ddict["detector_mode"] = ddict.pop("mode")
        ddict["detector_type"] = ddict.pop("type")
        ddict["detector_brightness"] = ddict.pop("brightness")
        ddict["detector_contrast"] = ddict.pop("contrast")
        ddict["current"] = ddict.pop("beam_current")

        return ddict
    
    @staticmethod
    def from_dict(settings: dict) -> 'BeamSystemSettings':
        return BeamSystemSettings(
            beam_type=BeamType[settings["beam_type"]],
            enabled=settings["enabled"],
            beam=BeamSettings.from_dict(settings),
            detector=FibsemDetectorSettings.from_dict(settings),
            eucentric_height=settings["eucentric_height"],
            column_tilt=settings["column_tilt"],
            plasma=settings.get("plasma", False),
            plasma_gas=settings.get("plasma_gas", None),
        )

@dataclass
class ManipulatorSystemSettings:
    enabled: bool
    rotation: bool
    tilt: bool

    def to_dict(self):
        return {
            "enabled": self.enabled,
            "rotation": self.rotation,
            "tilt": self.tilt,
        }
    
    @staticmethod
    def from_dict(settings: dict):
        return ManipulatorSystemSettings(
            enabled=settings["enabled"],
            rotation=settings["rotation"],
            tilt=settings["tilt"],
        )



@dataclass
class GISSystemSettings:
    enabled: bool
    multichem: bool
    sputter_coater: bool
    inserted: bool = False

    def to_dict(self):
        return {
            "enabled": self.enabled,
            "multichem": self.multichem,
            "sputter_coater": self.sputter_coater,
        }
    
    @staticmethod
    def from_dict(settings: dict):
        return GISSystemSettings(
            enabled=settings["enabled"],
            multichem=settings["multichem"],
            sputter_coater=settings["sputter_coater"],
        )

import fibsem

@dataclass
class SystemInfo:
    name: str
    ip_address: str
    manufacturer: str
    model: str
    serial_number: str
    hardware_version: str
    software_version: str
    fibsem_version: str = fibsem.__version__
    application: str = None
    application_version: str = None

    def to_dict(self):
        return {
            "name": self.name,
            "ip_address": self.ip_address,
            "manufacturer": self.manufacturer,
            "model": self.model,
            "serial_number": self.serial_number,
            "hardware_version": self.hardware_version,
            "software_version": self.software_version,
            "fibsem_version": self.fibsem_version,
            "application": self.application,
            "application_version": self.application_version,
        }
    
    @staticmethod
    def from_dict(settings: dict):
        return SystemInfo(
            name=settings.get("name", "Unknown"),
            ip_address=settings.get("ip_address", "Unknown"),
            manufacturer=settings.get("manufacturer", "Unknown"),
            model=settings.get("model", "Unknown"),
            serial_number=settings.get("serial_number", "Unknown"),
            hardware_version=settings.get("hardware_version", "Unknown"),
            software_version=settings.get("software_version", "Unknown"),
            fibsem_version=settings.get("fibsem_version", fibsem.__version__),
            application=settings.get("application", None),
            application_version=settings.get("application_version", None),  
        )

@dataclass
class SystemSettings:
    stage: StageSystemSettings
    electron: BeamSystemSettings
    ion: BeamSystemSettings
    manipulator: ManipulatorSystemSettings
    gis: GISSystemSettings
    info: SystemInfo    

    def to_dict(self):
        return {
            "stage": self.stage.to_dict(),
            "electron": self.electron.to_dict(),
            "ion": self.ion.to_dict(),
            "manipulator": self.manipulator.to_dict(),
            "gis": self.gis.to_dict(),
            "info": self.info.to_dict(),
        }
    
    @staticmethod
    def from_dict(settings: dict):

        # TODO: remove this once the settings are updated
        settings["electron"]["beam_type"] = BeamType.ELECTRON.name
        settings["ion"]["beam_type"] = BeamType.ION.name
            
        return SystemSettings(
            stage=StageSystemSettings.from_dict(settings["stage"]),
            electron=BeamSystemSettings.from_dict(settings["electron"]),
            ion=BeamSystemSettings.from_dict(settings["ion"]),
            manipulator=ManipulatorSystemSettings.from_dict(settings["manipulator"]),
            gis=GISSystemSettings.from_dict(settings["gis"]),
            info=SystemInfo.from_dict(settings["info"]),
        )

@dataclass
class MicroscopeSettings:

    """
    A data class representing the settings for a microscope system.

    Attributes:
        system (SystemSettings): An instance of the `SystemSettings` class that holds the system settings.
        image (ImageSettings): An instance of the `ImageSettings` class that holds the image settings.
        milling (FibsemMillingSettings): An instance of the `FibsemMillingSettings` class that holds the fibsem milling settings..
        protocol (dict, optional): A dictionary representing the protocol settings. Defaults to None.

    Methods:
        to_dict(): Returns a dictionary representation of the `MicroscopeSettings` object.
        from_dict(settings: dict, protocol: dict = None) -> "MicroscopeSettings": Returns an instance of the `MicroscopeSettings` class from a dictionary.
    """

    system: SystemSettings
    image: ImageSettings
    milling: FibsemMillingSettings
    protocol: dict = None

    def to_dict(self) -> dict:
        settings_dict = {
            "imaging": self.image.to_dict(),
            "protocol": self.protocol,
            "milling": self.milling.to_dict(),
        }
        settings_dict.update(self.system.to_dict())

        return settings_dict

    @staticmethod
    def from_dict(
        settings: dict, protocol: dict = None
    ) -> "MicroscopeSettings":
        
        if protocol is None:
            protocol = settings.get("protocol", {"name": "demo"})
     
        return MicroscopeSettings(
            system=SystemSettings.from_dict(settings),
            image=ImageSettings.from_dict(settings["imaging"]),
            protocol=protocol,
            milling=FibsemMillingSettings.from_dict(settings["milling"]),
        )





@dataclass
class FibsemExperiment:
    id: str = None
    method: str = None
    date: float = datetime.timestamp(datetime.now())
    application: str = "OpenFIBSEM"
    fibsem_version: str = fibsem.__version__
    application_version: str = None

    def to_dict(self) -> dict:
        """Converts to a dictionary."""
        return {
            "id": self.id,
            "method": self.method,
            "date": self.date,
            "application": self.application,
            "fibsem_version": self.fibsem_version,
            "application_version": self.application_version,
        }

    @staticmethod
    def from_dict(settings: dict) -> "FibsemExperiment":
        """Converts from a dictionary."""
        return FibsemExperiment(
            id=settings.get("id", "Unknown"),
            method=settings.get("method", "Unknown"),
            date=settings.get("date", "Unknown"),
            application=settings.get("application", "OpenFIBSEM"),
            fibsem_version=settings.get("fibsem_version", fibsem.__version__),
            application_version=settings.get("application_version", None),
        )


@dataclass
class FibsemUser:
    name: str = None
    email: str = None
    organization: str = None
    hostname: str = None
    # TODO: add host_ip_address

    def to_dict(self) -> dict:
        """Converts to a dictionary."""
        return {
            "name": self.name,
            "email": self.email,
            "organization": self.organization,
            "hostname": self.hostname,
        }

    @staticmethod
    def from_dict(settings: dict) -> "FibsemUser":
        """Converts from a dictionary."""
        return FibsemUser(
            name=settings.get("name", "Unknown"),
            email=settings.get("email", "Unknown"),
            organization=settings.get("organization", "Unknown"),
            hostname=settings.get("hostname", "Unknown"),
        )

    @staticmethod
    def from_environment() -> "FibsemUser":
        import platform
        import socket
        username = os.environ.get("USERNAME", "username")

        if platform.system() == "Windows":
            hostname = os.environ.get("COMPUTERNAME", "hostname")
        elif platform.system() in ["Linux", "Darwin"]:
            hostname = socket.gethostname()
        else:
            hostname = "hostname"
            
        user = FibsemUser(name=username, email="null", organization="null", hostname=hostname)
        
        return user


@dataclass
class FibsemImageMetadata:
    """Metadata for a FibsemImage."""

    image_settings: ImageSettings
    pixel_size: Point
    microscope_state: MicroscopeState
    system: SystemSettings = None
    version: str = METADATA_VERSION
    user: FibsemUser = FibsemUser()
    experiment: FibsemExperiment = FibsemExperiment()

    def to_dict(self) -> dict:
        """Converts metadata to a dictionary.

        Returns:
            dictionary: self as a dictionary
        """
        settings_dict = {}
        if self.image_settings is not None:
            settings_dict["image"] = self.image_settings.to_dict()
        if self.version is not None:
            settings_dict["version"] = self.version
        if self.pixel_size is not None:
            settings_dict["pixel_size"] = self.pixel_size.to_dict()
        if self.microscope_state is not None:
            settings_dict["microscope_state"] = self.microscope_state.to_dict()
            settings_dict["user"] = self.user.to_dict()
        settings_dict["experiment"] = self.experiment.to_dict()
        settings_dict["system"] = self.system.to_dict() if self.system is not None else {}

        return settings_dict

    @staticmethod
    def from_dict(settings: dict) -> "ImageSettings":
        """Converts a dictionary to metadata."""

        image_settings = ImageSettings.from_dict(settings["image"])
        version = settings.get("version", METADATA_VERSION)
        if settings["pixel_size"] is not None:
            pixel_size = Point.from_dict(settings["pixel_size"])
        if settings["microscope_state"] is not None:
            microscope_state = MicroscopeState.from_dict(
                settings["microscope_state"]
            )

        metadata = FibsemImageMetadata(
            image_settings=image_settings,
            version=version,
            pixel_size=pixel_size,
            microscope_state=microscope_state,
            # detector_settings=detector_settings, # TODO: remove this
            user=FibsemUser.from_dict(settings.get("user", {})),
            experiment=FibsemExperiment.from_dict(settings.get("experiment", {})),
            system=SystemSettings.from_dict(settings.get("system", {})),
        )
        return metadata

    if THERMO:
        # TODO: move to ImageSettings
        def image_settings_from_adorned(
            image=AdornedImage, beam_type: BeamType = BeamType.ELECTRON
        ) -> ImageSettings:
            from fibsem.utils import current_timestamp

            image_settings = ImageSettings(
                resolution=[image.width, image.height],
                dwell_time=image.metadata.scan_settings.dwell_time,
                hfw=image.width * image.metadata.binary_result.pixel_size.x,
                autocontrast=True,
                beam_type=beam_type,
                autogamma=True,
                save=False,
                path="path",
                filename=current_timestamp(),
                reduced_area=None,
            )
            return image_settings

    def compare_image_settings(self, image_settings: ImageSettings) -> bool:
        """Compares image settings to the metadata image settings.

        Args:
            image_settings (ImageSettings): Image settings to compare to.

        Returns:
            bool: True if the image settings match the metadata image settings.
        """
        assert (
            self.image_settings.resolution[0] == image_settings.resolution[0]
            and self.image_settings.resolution[1] == image_settings.resolution[1]
        ), f"resolution: {self.image_settings.resolution} != {image_settings.resolution}"
        assert (
            self.image_settings.dwell_time == image_settings.dwell_time
        ), f"dwell_time: {self.image_settings.dwell_time} != {image_settings.dwell_time}"
        assert (
            self.image_settings.hfw == image_settings.hfw
        ), f"hfw: {self.image_settings.hfw} != {image_settings.hfw}"
        assert (
            self.image_settings.autocontrast == image_settings.autocontrast
        ), f"autocontrast: {self.image_settings.autocontrast} != {image_settings.autocontrast}"
        assert (
            self.image_settings.beam_type.value == image_settings.beam_type.value
        ), f"beam_type: {self.image_settings.beam_type.value} != {image_settings.beam_type.value}"
        assert (
            self.image_settings.autogamma == image_settings.autogamma
        ), f"gamma: {self.image_settings.autogamma} != {image_settings.autogamma}"
        assert (
            self.image_settings.save == image_settings.save
        ), f"save: {self.image_settings.save} != {image_settings.save}"
        assert (
            self.image_settings.path == image_settings.path
        ), f"path: {self.image_settings.path} != {image_settings.path}"
        assert (
            self.image_settings.filename == image_settings.filename
        ), f"filename: {self.image_settings.filename} != {image_settings.filename}"
        assert (
            self.image_settings.reduced_area == image_settings.reduced_area
        ), f"reduced_area: {self.image_settings.reduced_area} != {image_settings.reduced_area}"

        return True


class FibsemImage:

    """
    Class representing a FibsemImage and its associated metadata.
    Has in built methods to deal with image types of TESCAN and ThermoFisher API

    Args:
        data (np.ndarray): The image data stored in a numpy array.
        metadata (FibsemImageMetadata, optional): The metadata associated with the image. Defaults to None.

    Methods:
        load(cls, tiff_path: str) -> "FibsemImage":
            Loads a FibsemImage from a tiff file.

            Args:
                tiff_path (path): path to the tif* file

            Returns:
                FibsemImage: instance of FibsemImage

        save(self, path: Path) -> None:
            Saves a FibsemImage to a tiff file.

            Inputs:
                path (path): path to save directory and filename
    """

    def __init__(self, data: np.ndarray, metadata: FibsemImageMetadata = None):
        if check_data_format(data):
            if data.ndim == 3 and data.shape[2] == 1:
                data = data[:, :, 0]
            self.data = data
        else:
            raise Exception("Invalid Data format for Fibsem Image")
        if metadata is not None:
            self.metadata = metadata
        else:
            self.metadata = None

    @classmethod
    def load(cls, tiff_path: str) -> "FibsemImage":
        """Loads a FibsemImage from a tiff file.

        Args:
            tiff_path (path): path to the tif* file

        Returns:
            FibsemImage: instance of FibsemImage
        """
        with tff.TiffFile(tiff_path) as tiff_image:
            data = tiff_image.asarray()
            try:
                metadata = json.loads(
                    tiff_image.pages[0].tags["ImageDescription"].value
                )
                metadata = FibsemImageMetadata.from_dict(metadata)
            except Exception as e:
                metadata = None
                # print(f"Error: {e}")
                # import traceback
                # traceback.print_exc()
        return cls(data=data, metadata=metadata)

    def save(self, path: Path = None) -> None:
        """Saves a FibsemImage to a tiff file.

        Inputs:
            path (path): path to save directory and filename
        """
        if path is None:
            path = os.path.join(
                self.metadata.image_settings.path,
                self.metadata.image_settings.filename,
            )
        os.makedirs(os.path.dirname(path), exist_ok=True)
        path = Path(path).with_suffix(".tif")

        if self.metadata is not None:
            metadata_dict = self.metadata.to_dict()
        else:
            metadata_dict = None
        tff.imwrite(
            path,
            self.data,
            metadata=metadata_dict,
        )

    if THERMO:

        @classmethod
        def fromAdornedImage(
            cls,
            adorned: AdornedImage,
            image_settings: ImageSettings,
            state: MicroscopeState = None,
        ) -> "FibsemImage":
            """Creates FibsemImage from an AdornedImage (microscope output format).

            Args:
                adorned (AdornedImage): Adorned Image from microscope
                metadata (FibsemImageMetadata, optional): metadata extracted from microscope output. Defaults to None.

            Returns:
                FibsemImage: instance of FibsemImage from AdornedImage
            """
            if state is None:
                state = MicroscopeState(
                    timestamp=adorned.metadata.acquisition.acquisition_datetime,
                    stage_position=FibsemStagePosition(
                        adorned.metadata.stage_settings.stage_position.x,
                        adorned.metadata.stage_settings.stage_position.y,
                        adorned.metadata.stage_settings.stage_position.z,
                        adorned.metadata.stage_settings.stage_position.r,
                        adorned.metadata.stage_settings.stage_position.t,
                    ),
                    electron_beam=BeamSettings(beam_type=BeamType.ELECTRON),
                    ion_beam=BeamSettings(beam_type=BeamType.ION),
                )
            else:
                state.timestamp = adorned.metadata.acquisition.acquisition_datetime

            pixel_size = Point(
                adorned.metadata.binary_result.pixel_size.x,
                adorned.metadata.binary_result.pixel_size.y,
            )

            metadata = FibsemImageMetadata(
                image_settings=image_settings,
                pixel_size=pixel_size,
                microscope_state=state,
            )
            return cls(data=adorned.data, metadata=metadata)

    if TESCAN:

        @classmethod
        def fromTescanImage(
            cls,
            image: Document,
            image_settings: ImageSettings,
            state: MicroscopeState,
            detector: FibsemDetectorSettings,
        ) -> "FibsemImage":
            """Creates FibsemImage from a tescan (microscope output format).

            Args:
                image (Tescan): Adorned Image from microscope
                metadata (FibsemImageMetadata, optional): metadata extracted from microscope output. Defaults to None.

            Returns:
                FibsemImage: instance of FibsemImage from AdornedImage
            """

            pixel_size = Point(
                float(image.Header["MAIN"]["PixelSizeX"]),
                float(image.Header["MAIN"]["PixelSizeY"]),
            )
            metadata = FibsemImageMetadata(
                image_settings=image_settings,
                pixel_size=pixel_size,
                microscope_state=state,
                # detector_settings=detector,
            )

            return cls(data=np.array(image.Image), metadata=metadata)

        @classmethod
        def fromTescanFile(
            cls,
            image_path: str,
            metadata_path: str,
            beam_type: BeamType,
        ) -> "FibsemImage":
            with tff.TiffFile(image_path) as tiff_image:
                data = tiff_image.asarray()

            stage = 0
            dictionary = {"MAIN": {}, "SEM": {}, "FIB": {}}
            with open(metadata_path, "r") as file:
                for line in file:
                    if line.startswith("["):
                        stage += 1
                        continue

                    line = line.strip()
                    if not line:
                        continue  # Skip empty lines

                    key, value = line.split("=")
                    key = key.strip()
                    value = value.strip()
                    if stage == 1:
                        dictionary["MAIN"][key] = value
                    if stage == 2 and beam_type.name == "ELECTRON":
                        dictionary["SEM"][key] = value
                    if stage == 2 and beam_type.name == "ION":
                        dictionary["FIB"][key] = value

            if beam_type.name == "ELECTRON":
                image_settings = ImageSettings(
                    resolution=[data.shape[0], data.shape[1]],
                    dwell_time=float(dictionary["SEM"]["DwellTime"]),
                    hfw=data.shape[0] * float(dictionary["MAIN"]["PixelSizeX"]),
                    beam_type=BeamType.ELECTRON,
                    filename=Path(image_path).stem,
                    path=Path(image_path).parent,
                )
                pixel_size = Point(
                    float(dictionary["MAIN"]["PixelSizeX"]),
                    float(dictionary["MAIN"]["PixelSizeY"]),
                )
                microscope_state = MicroscopeState(
                    timestamp=datetime.strptime(
                        dictionary["MAIN"]["Date"] + " " + dictionary["MAIN"]["Time"],
                        "%Y-%m-%d %H:%M:%S",
                    ),
                    electron_beam=BeamSettings(
                        beam_type=BeamType.ELECTRON,
                        working_distance=float(dictionary["SEM"]["WD"]),
                        beam_current=float(dictionary["SEM"]["PredictedBeamCurrent"]),
                        voltage=float(dictionary["SEM"]["TubeVoltage"]),
                        hfw=data.shape[0] * float(dictionary["MAIN"]["PixelSizeX"]),
                        resolution=[data.shape[0], data.shape[1]],
                        dwell_time=float(dictionary["SEM"]["DwellTime"]),
                        shift=Point(
                            float(dictionary["SEM"]["ImageShiftX"]),
                            float(dictionary["SEM"]["ImageShiftY"]),
                        ),
                        stigmation=Point(
                            float(dictionary["SEM"]["StigmatorX"]),
                            float(dictionary["SEM"]["StigmatorY"]),
                        ),
                    ),
                    ion_beam=BeamSettings(beam_type=BeamType.ION),
                )
                detector_settings = FibsemDetectorSettings(
                    type=dictionary["SEM"]["Detector"],
                    brightness=float(dictionary["SEM"]["Detector0Offset"]),
                    contrast=float(dictionary["SEM"]["Detector0Gain"]),
                )

            if beam_type.name == "ION":
                image_settings = ImageSettings(
                    resolution=[data.shape[0], data.shape[1]],
                    dwell_time=float(dictionary["FIB"]["DwellTime"]),
                    hfw=data.shape[0] * float(dictionary["MAIN"]["PixelSizeX"]),
                    beam_type=BeamType.ELECTRON,
                    filename=Path(image_path).stem,
                    path=Path(image_path).parent,
                )
                pixel_size = Point(
                    float(dictionary["MAIN"]["PixelSizeX"]),
                    float(dictionary["MAIN"]["PixelSizeY"]),
                )
                microscope_state = MicroscopeState(
                    timestamp=datetime.strptime(
                        dictionary["MAIN"]["Date"] + " " + dictionary["MAIN"]["Time"],
                        "%Y-%m-%d %H:%M:%S",
                    ),
                    electron_beam=BeamSettings(beam_type=BeamType.ELECTRON),
                    ion_beam=BeamSettings(
                        beam_type=BeamType.ION,
                        working_distance=float(dictionary["FIB"]["WD"]),
                        beam_current=float(dictionary["FIB"]["PredictedBeamCurrent"]),
                        hfw=data.shape[0] * float(dictionary["MAIN"]["PixelSizeX"]),
                        resolution=[data.shape[0], data.shape[1]],
                        dwell_time=float(dictionary["FIB"]["DwellTime"]),
                        shift=Point(
                            float(dictionary["FIB"]["ImageShiftX"]),
                            float(dictionary["FIB"]["ImageShiftY"]),
                        ),
                        stigmation=Point(
                            float(dictionary["FIB"]["StigmatorX"]),
                            float(dictionary["FIB"]["StigmatorY"]),
                        ),
                    ),
                )
                detector_settings = FibsemDetectorSettings(
                    type=dictionary["FIB"]["Detector"],
                    brightness=float(dictionary["FIB"]["Detector0Offset"]) / 100,
                    contrast=float(dictionary["FIB"]["Detector0Gain"]) / 100,
                )

            metadata = FibsemImageMetadata(
                image_settings=image_settings,
                pixel_size=pixel_size,
                microscope_state=microscope_state,
                # detector_settings=detector_settings,
                version=METADATA_VERSION,
            )
            return cls(data=data, metadata=metadata)


@dataclass
class ReferenceImages:
    low_res_eb: FibsemImage
    high_res_eb: FibsemImage
    low_res_ib: FibsemImage
    high_res_ib: FibsemImage

    def __iter__(self) -> list[FibsemImage]:
        yield self.low_res_eb, self.high_res_eb, self.low_res_ib, self.high_res_ib


class ThermoGISLine:
    def __init__(self, line=None, name=None, status: str = "Retracted"):
        self.line = line
        self.name = name
        self.status = status
        self.temp_ready = False

    def insert(self):
        if self.line is not None:
            self.line.insert()
        self.status = "Inserted"

    def retract(self):
        if self.line is not None:
            self.line.retract()
        self.status = "Retracted"


class ThermoMultiChemLine:
    def __init__(self, line=None, status: str = "Retracted"):
        self.line = line
        self.status = status
        self.positions = ["Electron Default", "Ion Default", "Retract"]
        self.current_position = "Retract"
        self.temp_ready = False

    def insert(self, position):
        # position_str = getattr(MultiChemInsertPosition,position)

        if self.line is not None:
            self.line.insert(position)

        self.current_position = position
        self.status = "Inserted"

    def retract(self):
        if self.line is not None:
            self.line.retract()

        self.status = "Retracted"
        self.current_position = "Retracted"


def check_data_format(data: np.ndarray) -> bool:
    """Checks that data is in the correct format."""
    # assert data.ndim == 2  # or data.ndim == 3
    # assert data.dtype in [np.uint8, np.uint16]
    if data.ndim == 3 and data.shape[2] == 1:
        data = data[:, :, 0]
    return data.ndim == 2 and data.dtype in [np.uint8, np.uint16]
